# -*- coding: utf-8 -*-

import logging

from time import time

import numpy as np

from collections import OrderedDict

import pyhrf
from pyhrf.ndarray import xndarray
from pyhrf.vbjde.vem_bold_models_fast_ms import Main_vbjde_physio
from pyhrf.vbjde.vem_tools import roc_curve
from pyhrf.xmlio import XmlInitable
from pyhrf.tools import format_duration
from pyhrf.ui.jde import JDEAnalyser
from pyhrf.sandbox.physio_params import PHY_PARAMS_KHALIDOV11


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def change_dim(labels):
    '''
    Change labels dimension from
    (ncond, nclass, nvox)
    to
    (nclass, ncond, nvox)
    '''
    ncond = labels.shape[0]
    nclass = labels.shape[1]
    nvox = labels.shape[2]
    newlabels = np.zeros((nclass, ncond, nvox))
    for cond in xrange(ncond):
        for clas in xrange(nclass):
            for vox in xrange(nvox):
                newlabels[clas][cond][vox] = labels[cond][clas][vox]

    return newlabels


class JDEVEMAnalyser(JDEAnalyser):

    parametersComments = {
        'dt': 'time resolution of the estimated HRF in seconds',
        'hrfDuration': 'duration of the HRF in seconds',
        'fast': 'running fast VEM with C extensions',
        'nbClasses': 'number of classes for the response levels',
        'PLOT': 'plotting flag for convergence curves',
        'nItMax': 'maximum iteration number',
        'nItMin': 'minimum iteration number',
        'scale': 'flag for the scaling factor applied to the data fidelity '
        'term during m_h step.\n'
        'If scale=False then do nothing, else divide '
        'the data fidelity term by the number of voxels',
        'beta': 'initial value of spatial Potts regularization parameter',
        'simulation': 'indicates whether the run corresponds to a simulation \
                        example or not',
        'estimateSigmaH': 'estimate or not the HRF variance',
        'estimateSigmaG': 'estimate or not the PRF variance',
        'estimateH': 'estimate or not the HRF',
        'estimateG': 'estimate or not the PRF',
        'sigmaH': 'Initial HRF variance',
        'sigmaG': 'Initial PRF variance',
        'estimateBeta': 'estimate or not the Potts spatial regularization '
        'parameter',
        'estimateLA': 'Explicit drift and perfusion baseline estimation',
        'estimateLabels': 'estimate or not the Labels',
        'estimateMixtParam': 'estimate or not the mixture parameters',
        'InitVar': 'Initiale value of active and inactive gaussian variances',
        'InitMean': 'Initiale value of active gaussian means',
        'constrained': 'adding constrains: positivity and norm = 1 ',
        'zero_constrained': 'putting first and last point of the HRF to zero '
        }

    parametersToShow = ['dt', 'hrfDuration', 'nItMax', 'nItMin',
                        'estimateSigmaH', 'estimateSigmaG', 'estimateH',
                        'estimateG', 'estimateBeta',
                        'estimateLabels', 'estimateLA',
                        'estimateMixtParam', 'InitVar', 'InitMean',
                        'scale', 'nbClasses', 'fast', 'PLOT',
                        'sigmaH', 'sigmaG']

    def __init__(self, hrfDuration=25., dt=.5, fast=True, constrained=False,
                 nbClasses=2, PLOT=False, nItMax=1, nItMin=1, scale=False,
                 beta=1.0, simulation=None, fmri_data=None, computeContrast=True,
                 estimateH=True, estimateG=True, use_hyperprior=False,
                 estimateSigmaH=True, estimateSigmaG=True, positivity=False,
                 sigmaH=0.0001, sigmaG=0.0001, sigmaMu=0.0001, physio=True,
                 gammaH=1000, gammaG=1000, zero_constrained=False,
                 estimateLabels=True, estimateMixtParam=True, contrasts=None,
                 InitVar=0.5, InitMean=2.0, estimateA=True, estimateC=True,
                 estimateBeta=True, estimateNoise=True, estimateLA=True,
                 phy_params=PHY_PARAMS_KHALIDOV11, prior='no', n_session=1):

        XmlInitable.__init__(self)
        JDEAnalyser.__init__(self, outputPrefix='jde_vem_asl_')

        # Important thing : all parameters must have default values
        self.hrfDuration = hrfDuration
        self.dt = dt
        self.fast = fast
        self.constrained = constrained
        self.nbClasses = nbClasses
        self.PLOT = PLOT
        self.nItMax = nItMax
        self.nItMin = nItMin
        self.scale = scale
        self.beta = beta
        self.simulation = simulation
        self.fmri_data = fmri_data
        self.estimateH = estimateH
        self.estimateG = estimateG
        self.estimateSigmaH = estimateSigmaH
        self.estimateSigmaG = estimateSigmaG
        self.sigmaH = sigmaH
        self.sigmaG = sigmaG
        self.sigmaMu = sigmaMu
        self.gammaH = gammaH
        self.gammaG = gammaG
        self.estimateLabels = estimateLabels
        self.estimateMixtParam = estimateMixtParam
        self.InitVar = InitVar
        self.InitMean = InitMean
        self.estimateA = estimateA
        self.estimateC = estimateC
        self.estimateBeta = estimateBeta
        self.estimateNoise = estimateNoise
        self.estimateLA = estimateLA
        self.use_hyperprior = use_hyperprior
        self.positivity = positivity
        self.physio = physio
        self.prior = prior
        if contrasts is None:
            contrasts = OrderedDict()
        self.contrasts = contrasts
        self.computeContrast = computeContrast
        self.phy_params = phy_params
        self.n_session = n_session
        self.zc = zero_constrained

        logger.info("VEM analyzer:")
        logger.info(" - estimate sigma H: %s", str(self.estimateSigmaH))
        logger.info(" - init sigma H: %f", self.sigmaH)
        logger.info(" - estimate drift and perfusion baseline: %s",
                    str(self.estimateLA))

    def analyse_roi(self, roiData):
        # roiData is of type FmriRoiData, see pyhrf.core.FmriRoiData
        # roiData.bold : numpy array of shape
        # BOLD has shape (nscans, nvoxels)
        # roiData.graph #list of neighbours
        n_scan_allsession, nvox = roiData.bold.shape
        n_scan = n_scan_allsession / self.n_session
        data0 = roiData.bold.reshape(self.n_session, n_scan, nvox)
        data = np.zeros_like(data0)
        for s in xrange(self.n_session):
            data_mean = np.mean(data0[s, :, :])
            data_range = (np.max(data0[s, :, :]) - np.min(data0[s, :, :]))
            data[s, :, :] = (data0[s, :, :] - data_mean) * 100 / data_range
        Onsets = roiData.paradigm.get_joined_onsets_dim()
        durations = roiData.paradigm.get_joined_durations_dim()
        TR = roiData.tr
        beta = self.beta
        scale = 1                   # roiData.nbVoxels
        #nvox = roiData.get_nb_vox_in_mask()
        if self.scale:
            scale = nvox
        rid = roiData.get_roi_id()
        logger.info("JDE VEM - roi %d, nvox=%d, nconds=%d, nItMax=%d", rid,
                    nvox, len(Onsets), self.nItMax)

        #self.contrasts.pop('dummy_example', None)
        cNames = roiData.paradigm.get_stimulus_names()
        graph = roiData.get_graph()
        idx_tag1 = roiData.get_extra_data('asl_first_tag_scan_idx', 0)

        t_start = time()

        logger.info("fast VEM with drift estimation and a constraint")
        try:
            simu = roiData.simulation[0]
        except:
            try:
                simu = roiData.simulation
            except:
                simu = None

        if self.physio:
            NbIter, brls, estimated_brf, prls, estimated_prf, labels, \
            noiseVar, mu_Ma, sigma_Ma, mu_Mc, sigma_Mc, Beta, L, PL, alpha,\
            Sigma_brls, Sigma_prls, Sigma_brf, Sigma_prf, rerror, \
            CONTRAST_A, CONTRASTVAR_A, CONTRAST_C, CONTRASTVAR_C, \
            cA, cH, cC, cG, cZ, cAH, cCG, cTime, FE = Main_vbjde_physio(
                                       graph, data, Onsets, durations, self.hrfDuration,
                                       self.nbClasses, TR, beta, self.dt, scale=scale,
                                       estimateSigmaG=self.estimateSigmaG,
                                       sigmaH=self.sigmaH, sigmaG=self.sigmaG,
                                       gamma_h=self.gammaH, gamma_g=self.gammaG,
                                       NitMax=self.nItMax, NitMin=self.nItMin,
                                       estimateSigmaH=self.estimateSigmaH,
                                       estimateBeta=self.estimateBeta, PLOT=self.PLOT,
                                       contrasts=self.contrasts,
                                       computeContrast=self.computeContrast,
                                       idx_first_tag=idx_tag1,
                                       simulation=simu, sigmaMu=self.sigmaMu,
                                       estimateH=self.estimateH,
                                       estimateG=self.estimateG,
                                       estimateA=self.estimateA,
                                       estimateC=self.estimateC,
                                       estimateNoise=self.estimateNoise,
                                       estimateMP=self.estimateMixtParam,
                                       estimateZ=self.estimateLabels,
                                       estimateLA=self.estimateLA,
                                       constraint=self.constrained,
                                       positivity=self.positivity,
                                       use_hyperprior=self.use_hyperprior,
                                       phy_params=self.phy_params,
                                       prior=self.prior, zc=self.zc)

        # Plot analysis duration
        self.analysis_duration = time() - t_start
        logger.info('JDE VEM analysis took: %s',
                    format_duration(self.analysis_duration))

        # OUTPUTS: Pack all outputs within a dict
        logger.info("Preparing outputs... ")
        outputs = {}
        brf_time = np.arange(len(estimated_brf)) * self.dt
        outputs['brf'] = xndarray(estimated_brf, axes_names=['time'],
                                  axes_domains={'time': brf_time},
                                  value_label="BRF")
        #logger.info("BRF prepared ")
        domCondition = {'condition': cNames}
        outputs['brls'] = xndarray(brls.T, value_label="BRLs",
                                   axes_names=['condition', 'voxel'],
                                   axes_domains=domCondition)
        #logger.info("BRLs prepared ")
        prf_time = np.arange(len(estimated_prf)) * self.dt
        outputs['prf'] = xndarray(estimated_prf, axes_names=['time'],
                                  axes_domains={'time': prf_time},
                                  value_label="PRF")
        #logger.info("PRF prepared ")
        outputs['prls'] = xndarray(prls.T, value_label="PRLs",
                                   axes_names=['condition', 'voxel'],
                                   axes_domains=domCondition)
        #logger.info("PRLs prepared ")

        outputs['Sigma_brf'] = xndarray(Sigma_brf, value_label="Sigma_BRF")
        #logger.info("Sigma_BRF prepared ")
        outputs['Sigma_prf'] = xndarray(Sigma_prf, value_label="Sigma_PRF")
        #logger.info("Sigma_PRF prepared ")

        ad = {'condition': cNames, 'condition2': Onsets.keys()}
        outputs['Sigma_brls'] = xndarray(Sigma_brls, value_label="Sigma_BRLs",
                                         axes_names=['condition', 'condition2',
                                                     'voxel'],
                                         axes_domains=ad)
        #logger.info("Sigma_a prepared ")
        outputs['Sigma_prls'] = xndarray(Sigma_prls, value_label="Sigma_PRLs",
                                         axes_names=['condition', 'condition2',
                                                     'voxel'],
                                         axes_domains=ad)
        #logger.info("Sigma_c prepared ")
        outputs['NbIter'] = xndarray(np.array([NbIter]), value_label="NbIter")
        outputs['beta'] = xndarray(Beta, value_label="beta",
                                   axes_names=['condition'],
                                   axes_domains=domCondition)

        #logger.info("perfusion baseline prepared ")
        outputs['alpha'] = xndarray(alpha, value_label="Perf_baseline",
                                          axes_names=['voxel'])

        #logger.info("Beta prepared ")
        nbc, nbv = len(cNames), brls.shape[0]
        repeatedBeta = np.repeat(Beta, nbv).reshape(nbc, nbv)
        outputs['beta_mapped'] = xndarray(repeatedBeta, value_label="beta",
                                          axes_names=['condition', 'voxel'],
                                          axes_domains=domCondition)

        repeated_brf = np.repeat(estimated_brf, nbv).reshape(-1, nbv)
        outputs["brf_mapped"] = xndarray(repeated_brf, value_label="BRFs",
                                         axes_names=["time", "voxel"],
                                         axes_domains={"time": brf_time})

        repeated_prf = np.repeat(estimated_prf, nbv).reshape(-1, nbv)
        outputs["prf_mapped"] = xndarray(repeated_prf, value_label="PRFs",
                                         axes_names=["time", "voxel"],
                                         axes_domains={"time": prf_time})

        #logger.info("beta mapped prepared ")
        outputs['roi_mask'] = xndarray(np.zeros(nbv) + roiData.get_roi_id(),
                                       value_label="ROI",
                                       axes_names=['voxel'])

        #logger.info("ROI mask prepared ")
        mixtpB = np.zeros((roiData.nbConditions, self.nbClasses, 2))
        mixtpB[:, :, 0] = mu_Ma
        mixtpB[:, :, 1] = sigma_Ma ** 2
        mixtpP = np.zeros((roiData.nbConditions, self.nbClasses, 2))
        mixtpP[:, :, 0] = mu_Mc
        mixtpP[:, :, 1] = sigma_Mc ** 2
        an = ['condition', 'Act_class', 'component']
        ad = {'Act_class': ['inactiv', 'activ'],
              'condition': cNames,
              'component': ['mean', 'var']}
        outputs['mixt_pB'] = xndarray(mixtpB, axes_names=an, axes_domains=ad)
        outputs['mixt_pP'] = xndarray(mixtpP, axes_names=an, axes_domains=ad)
        #logger.info("Mixture parameters prepared ")
        an = ['condition', 'Act_class', 'voxel']
        ad = {'Act_class': ['inactiv', 'activ'],
              'condition': cNames}
        #logger.info("mixt params prepared ")
        outputs['labels'] = xndarray(labels, value_label="Labels",
                                     axes_names=an, axes_domains=ad)
        #logger.info("labels prepared ")
        outputs['noiseVar'] = xndarray(noiseVar, value_label="noiseVar",
                                       axes_names=['voxel'])
        #logger.info("noise variance prepared ")
        if self.estimateLA:
            outputs['drift_coeff'] = xndarray(L, value_label="Drift",
                                              axes_names=['coeff', 'voxel'])
            outputs['drift'] = xndarray(PL, value_label="Delta BOLD",
                                        axes_names=['time', 'voxel'])
            logger.info("drift prepared ")
        logger.info("outputs prepared ")

        if (len(self.contrasts) >0) and self.computeContrast:
            #keys = list((self.contrasts[nc]) for nc in self.contrasts)
            domContrast = {'contrast':self.contrasts.keys()}
            outputs['contrastsA'] = xndarray(CONTRAST_A, value_label="Contrast_A",
                                            axes_names=['voxel','contrast'],
                                            axes_domains=domContrast)
            outputs['contrastsC'] = xndarray(CONTRAST_C, value_label="Contrast_C",
                                            axes_names=['voxel','contrast'],
                                            axes_domains=domContrast)
            c = xndarray(CONTRASTVAR_A, value_label="Contrasts_Variance_A",
                         axes_names=['voxel','contrast'],
                         axes_domains=domContrast)
            outputs['contrasts_variance_a'] = c
            outputs['ncontrasts_a'] = xndarray(CONTRAST_A/CONTRASTVAR_A**.5,
                                             value_label="Normalized Contrast A",
                                             axes_names=['voxel','contrast'],
                                             axes_domains=domContrast)
            c = xndarray(CONTRASTVAR_C, value_label="Contrasts_Variance_C",
                         axes_names=['voxel','contrast'],
                         axes_domains=domContrast)
            outputs['contrasts_variance_c'] = c
            outputs['ncontrasts_c'] = xndarray(CONTRAST_C/CONTRASTVAR_C**.5,
                                             value_label="Normalized Contrast C",
                                             axes_names=['voxel','contrast'],
                                             axes_domains=domContrast)

        #######################################################################
        # CONVERGENCE
        if 1:
            cTimeMean = cTime[-1] / np.float(NbIter)
            logger.info("Saving convergence... ")

            axes_names = ['duration']
            ax = (np.arange(self.nItMax) + 1) * cTimeMean
            ax[:len(cTime)] = cTime
            ad = {'duration': ax}

            outName = 'convergence_Labels'
            c = np.zeros(self.nItMax)  # -.001 #
            c[:len(cZ)] = cZ
            outputs[outName] = xndarray(c, axes_names=axes_names,
                                        axes_domains=ad,
                                        value_label='Conv_Criterion_Z')
            outName = 'convergence_BRF'
            #ad = {'Conv_Criterion':np.arange(len(cH))}
            c = np.zeros(self.nItMax)   # -.001 #
            c[:len(cH)] = cH
            outputs[outName] = xndarray(c, axes_names=axes_names,
                                        axes_domains=ad,
                                        value_label='Conv_Criterion_H')
            outName = 'convergence_BRL'
            c = np.zeros(self.nItMax)  # -.001 #
            c[:len(cA)] = cA
            #ad = {'Conv_Criterion':np.arange(len(cA))}
            outputs[outName] = xndarray(c, axes_names=axes_names,
                                        axes_domains=ad,
                                        value_label='Conv_Criterion_A')
            outName = 'convergence_PRF'
            #ad = {'Conv_Criterion':np.arange(len(cH))}
            c = np.zeros(self.nItMax)   # -.001 #
            c[:len(cG)] = cG
            outputs[outName] = xndarray(c, axes_names=axes_names,
                                        axes_domains=ad,
                                        value_label='Conv_Criterion_G')
            outName = 'convergence_PRL'
            c = np.zeros(self.nItMax)  # -.001 #
            c[:len(cC)] = cC
            #ad = {'Conv_Criterion':np.arange(len(cA))}
            outputs[outName] = xndarray(c, axes_names=axes_names,
                                        axes_domains=ad,
                                        value_label='Conv_Criterion_C')
            outName = 'convergence_FE'
            c = np.zeros(self.nItMax)  # -.001 #
            c[:len(FE)] = FE
            outputs[outName] = xndarray(c, axes_names=axes_names,
                                        axes_domains=ad,
                                        value_label='Conv_Criterion_FE')
            logger.info("Convergence saved ")

        #######################################################################
        # SIMULATION
        if self.simulation is not None and 0:
            logger.info("Prepare parameters to compare if simulation")
            M = labels.shape[0]
            K = labels.shape[1]
            J = labels.shape[2]

            true_labels = np.zeros((M, J))
            for m in xrange(0,M):
                true_labels[m, :] = roiData.simulation[0]['labels'][m].flatten()

            newlabels = np.reshape(labels[:, 1, :], (M, J))
            #true_labels = roiData.simulation[0]['labels']
            #newlabels = labels

            se = []
            sp = []
            size = np.prod(labels.shape)
            for i in xrange(0, 2):  # (0, M):
                se0, sp0, auc = roc_curve(newlabels[i, :].tolist(),
                                          true_labels[i, :].tolist())
                se.append(se0)
                sp.append(sp0)
                size = min(size, len(sp0))
            SE = np.zeros((M, size), dtype=float)
            SP = np.zeros((M, size), dtype=float)
            for i in xrange(0, 2):  # M):
                tmp = np.array(se[i])
                SE[i, :] = tmp[0:size]
                tmp = np.array(sp[i])
                SP[i, :] = tmp[0:size]
            sensData, specData = SE, SP
            axes_names = ['1-specificity', 'condition']
            outName = 'ROC_audio'
            #ad = {'1-specificity': specData[0], 'condition': cNames}
            outputs[outName] = xndarray(sensData, axes_names=axes_names,
                                        #axes_domains=ad,
                                        value_label='sensitivity')

            m = specData[0].min()
            import matplotlib.font_manager as fm
            import matplotlib.pyplot as plt
            plt.figure(200)
            plt.plot(sensData[0], specData[0], '--', color='k', linewidth=2.0,
                     label='m=1')
            plt.hold(True)
            plt.plot(sensData[1], specData[1], color='k', linewidth=2.0,
                     label='m=2')
            # legend(('audio','video'))
            plt.xticks(color='k', size=14, fontweight='bold')
            plt.yticks(color='k', size=14, fontweight='bold')
            #xlabel('1 - Specificity',fontsize=16,fontweight='bold')
            # ylabel('Sensitivity',fontsize=16,fontweight='bold')
            prop = fm.FontProperties(size=14, weight='bold')
            plt.legend(loc=1, prop=prop)
            plt.axis([0., 1., m, 1.02])

            true_labels = roiData.simulation[0]['labels']
            true_brls = roiData.simulation[0]['nrls']
            true_prls = roiData.simulation[0]['prls']
            true_brf = roiData.simulation[0]['hrf'][:, 0]
            true_prf = roiData.simulation[0]['prf'][:, 0]
            true_drift = roiData.simulation[0]['drift']
            true_noise = roiData.simulation[0]['noise']

            if simu is not None:
                logger.info("Check parameters errors")
                self.finalizeEstimation(true_labels, newlabels, nvox,
                                        true_brf, estimated_brf,
                                        true_prf, estimated_prf,
                                        true_brls, brls.T,
                                        true_prls, prls.T,
                                        true_drift, PL, L,
                                        true_noise, noiseVar)

        # END SIMULATION
        #######################################################################
        d = {'parcel_size': np.array([nvox])}
        outputs['analysis_duration'] = xndarray(np.array(
                                                [self.analysis_duration]),
                                                axes_names=['parcel_size'],
                                                axes_domains=d)
        """outputs['rerror'] = xndarray(np.array(  rerror),
                                                axes_names=['parcel_size'])"""
        return outputs


    def finalizeEstimation(self, true_labels, labels, nvox,
                           true_brf, estimated_brf, true_prf, estimated_prf,
                           true_brls, brls, true_prls, prls, true_drift, PL, L,
                           true_noise, noise):
        msg = []
        tol = .1
        fv = PL
        tv = true_drift
        delta = np.abs((fv - tv) / np.maximum(tv, fv))
        crit = (delta > tol).any()
        if crit:
            m = "Final value of %s is not close to " \
                "true value (mean delta=%f).\n" \
                " Final value:\n %s\n True value:\n %s\n" \
                % ('drift', delta.mean(), str(fv), str(tv))
            msg.append(m)

        delta = (((estimated_prf - true_prf) ** 2).sum() /
                 (true_prf ** 2).sum()) ** .5
        tol = 0.05
        crit = (delta > tol).any()
        if crit:
            m = "Final value of %s is not close to " \
                "true value (mean delta=%f).\n" \
                " Final value:\n %s\n True value:\n %s\n" \
                % ('prf', delta.mean(), str(estimated_prf), str(true_prf))
            msg.append(m)

        delta = (((estimated_brf - true_brf) ** 2).sum() /
                 (true_brf ** 2).sum()) ** .5
        tol = 0.05
        crit = (delta > tol).any()
        if crit:
            m = "Final value of %s is not close to " \
                "true value (mean delta=%f).\n" \
                " Final value:\n %s\n True value:\n %s\n" \
                % ('brf', delta.mean(), str(estimated_brf), str(true_brf))
            msg.append(m)

        tol = .1
        delta = ((labels != true_labels) * np.ones(labels.shape)).sum() / nvox
        crit = (delta > tol).any()
        if crit:
            m = "Final value of %s is not close to " \
                "true value (mean delta=%f).\n" \
                " Final value:\n %s\n True value:\n %s\n" \
                % ('labels', delta.mean(), str(labels), str(true_labels))
            msg.append(m)

        delta = np.abs((brls - true_brls) / np.maximum(true_brls, brls))
        crit = (delta > tol).any()
        if crit:
            m = "Final value of %s is not close to " \
                "true value (mean delta=%f).\n" \
                " Final value:\n %s\n True value:\n %s\n" \
                % ('brls', delta.mean(), str(brls), str(true_brls))
            msg.append(m)

        delta = np.abs((prls - true_prls) / np.maximum(true_prls, prls))
        crit = (delta > tol).any()
        if crit:
            m = "Final value of %s is not close to " \
                "true value (mean delta=%f).\n" \
                " Final value:\n %s\n True value:\n %s\n" \
                % ('prls', delta.mean(), str(prls), str(true_prls))
            msg.append(m)

        true_vnoise = np.var(true_noise, 0)
        delta = np.abs((noise - true_vnoise) / np.maximum(true_vnoise, noise))
        crit = (delta > tol).any()
        if crit:
            m = "Final value of %s is not close to " \
                "true value (mean delta=%f).\n" \
                " Final value:\n %s\n True value:\n %s\n" \
                % ('noise', delta.mean(), str(noise), str(true_vnoise))
            msg.append(m)
        print "\n".join(msg)


# Function to use directly in parallel computation
def run_analysis(**params):
    # from pyhrf.ui.vb_jde_analyser import JDEVEMAnalyser
    # import pyhrf
    # pyhrf.verbose.set_verbosity(1)
    # pyhrf.logger.setLevel(logging.INFO)
    fdata = params.pop('roi_data')
    # print 'doing params:'
    # print params
    vem_analyser = JDEVEMAnalyser(**params)
    return (dict([('ROI', fdata.get_roi_id())] + params.items()),
            vem_analyser.analyse_roi(fdata))



# -*- coding: utf-8 -*-

import os.path as op
import os
import sys
import traceback
import StringIO
import logging

import numpy as np

from pyhrf import xmlio, FmriData, FmriGroupData
from pyhrf.ndarray import MRI3Daxes
from pyhrf.tools import stack_trees, add_prefix
from pyhrf.tools._io import read_volume, read_texture
from pyhrf.ndarray import stack_cuboids


logger = logging.getLogger(__name__)


class FMRIAnalyser(xmlio.XmlInitable):

    P_OUTPUT_PREFIX = 'outputPrefix'
    P_ROI_AVERAGE = 'averageRoiBold'

    parametersToShow = [P_ROI_AVERAGE, P_OUTPUT_PREFIX]

    parametersComments = {
        P_ROI_AVERAGE: 'Average BOLD signals within each ROI before analysis.',
        P_OUTPUT_PREFIX: 'Tag to prefix every output name',
    }

    def __init__(self, outputPrefix='', roiAverage=False, pass_error=True,
                 gzip_outputs=False):
        xmlio.XmlInitable.__init__(self)
        if len(outputPrefix) == 0:
            outputPrefix = 'pyhrf_'
        self.outPrefix = outputPrefix
        self.outFile = add_prefix('outputs.xml', self.outPrefix)
        self.roiAverage = roiAverage
        self.pass_error = pass_error
        self.gzip_outputs = gzip_outputs

    def get_label(self):
        return 'pyhrf_fmri_analysis'

    def set_pass_errors(self, pass_error):
        self.pass_error = pass_error

    def set_gzip_outputs(self, gzip_outputs):
        self.gzip_outputs = gzip_outputs

    def __call__(self, *args, **kargs):
        return self.analyse_roi_wrap(*args, **kargs)

    def split_data(self, fdata, output_dir=None):

        if self.roiAverage:
            logger.info('Averaging ROI ...')
            fdata.average()  # TODO : debug sampling in JDE when averaging ...
        else:
            logger.info('Explode data ...')

        fdata.build_graphs()
        return fdata.roi_split()

    def analyse_roi_wrap_bak(self, roiData):

        report = 'ok'
        try:
            res = self.analyse_roi(roiData)
        except Exception:
            logger.error('!! Sampling crashed !!')
            logger.error('Exception traceback :')
            sio = StringIO.StringIO()
            traceback.print_exc(file=sio)
            # Make sure that the traceback is well garbage collected and
            # won't keep hidden references to objects :
            sys.traceback = None
            sys.exc_traceback = None
            report = sio.getvalue()
            sio.close()
            del sio
            logger.error(report)
            res = None

        return (roiData, res, report)

    def analyse_roi_wrap(self, roiData):
        """
        Wrap the analyse_roi method to catch potential exception
        """
        report = 'ok'
        if self.pass_error:
            try:
                res = self.analyse_roi(roiData)
            except Exception:
                logger.error('!! Sampling crashed !!')
                logger.error('Exception traceback :')
                sio = StringIO.StringIO()
                traceback.print_exc(file=sio)
                # Make sure that the traceback is well garbage collected and
                # won't keep hidden references to objects :
                sys.traceback = None
                sys.exc_traceback = None
                report = sio.getvalue()
                sio.close()
                del sio
                logger.error(report)
                res = None
        else:
            res = self.analyse_roi(roiData)

        return (roiData, res, report)

    def analyse_roi(self, roiData):
        raise NotImplementedError('%s does not implement roi analysis.'
                                  % self.__class__)

    def analyse(self, data, output_dir=None):
        """
        Launch the wrapped analyser onto the given data

        Args:
            data (~pyhrf.core.FmriData): the input fMRI data set (there may be multi parcels)
            output_dir (str): the path where to store parcel-specific fMRI data sets (after splitting according to
                the parcellation mask)

        Return:
            a list of analysis results
            ->  (list of tuple(FmriData, None|output of analyse_roi, str))
            =   (list of tuple(parcel data, analysis results, analysis report))
            See method analyse_roi_wrap
        """
        logger.info("Split data ...")
        explodedData = self.split_data(data, output_dir)
        logger.info("Data splitting returned %d rois", len(explodedData))
        return [self.analyse_roi_wrap(d) for d in explodedData]

    def filter_crashed_results(self, results):
        to_pop = []
        for i, r in enumerate(results[:]):
            roi_data, result, report = r
            roi_id = roi_data.get_roi_id()
            if report != 'ok':
                logger.error('-> Sampling crashed, roi %d!', roi_id)
                logger.error(report)
                to_pop.insert(0, i)

            elif result is None:
                logger.error('-> Sampling crashed (result is None), roi %d!',
                            roi_id)
                to_pop.insert(0, i)

        for i in to_pop:
            results.pop(i)

        return results

    def make_outputs_single_subject(self, data_rois, irois, all_outputs,
                                    targetAxes, ext, meta_data, output_dir):

        coutputs = {}
        output_fns = []

        roi_masks = [(data_roi.roiMask != data_roi.backgroundLabel)
                     for data_roi in data_rois]
        np_roi_masks = [np.where(roi_mask) for roi_mask in roi_masks]

        for output_name, roi_outputs in all_outputs.iteritems():
            logger.info('Merge output %s ...', output_name)
            try:
                if roi_outputs[0].has_axis('voxel'):
                    logger.debug('Merge as expansion ...')
                    dest_c = None
                    for i_roi, c in enumerate(roi_outputs):
                        dest_c = c.expand(roi_masks[i_roi],
                                          'voxel', targetAxes,
                                          dest=dest_c, do_checks=False,
                                          m=np_roi_masks[i_roi])
                else:
                    logger.debug('Merge as stack (%d elements)...', len(irois))
                    c_to_stack = [roi_outputs[i] for i in np.argsort(irois)]
                    # print 'c_to_stack:', c_to_stack
                    # print 'sorted(irois):', sorted(irois)
                    dest_c = stack_cuboids(c_to_stack, domain=sorted(irois),
                                           axis='ROI')
            except Exception, e:
                print "Could not merge outputs for %s" % output_name
                print "Exception was:"
                print e
                # raise e #stop here
            if 0 and dest_c is not None:
                print '-> ', dest_c.data.shape
            output_fn = op.join(output_dir, output_name + ext)
            output_fn = add_prefix(output_fn, self.outPrefix)
            output_fns.append(output_fn)
            logger.debug('Save output %s to %s', output_name, output_fn)
            try:
                if dest_c.meta_data:
                    tmp_meta_data = (meta_data[0], meta_data[1].copy())
                    tmp_meta_data[1]["descrip"] = dest_c.meta_data[1]["descrip"]
                    dest_c.meta_data = tmp_meta_data
                    dest_c.save(output_fn, set_MRI_orientation=True)
                else:
                    dest_c.save(output_fn, meta_data=meta_data,
                                set_MRI_orientation=True)
            except Exception:
                print 'Could not save output "%s", error stack was:' \
                    % output_name
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_exception(exc_type, exc_value, exc_traceback,
                                          limit=4, file=sys.stdout)
            coutputs[output_name] = dest_c

        return coutputs, output_fns

    def make_outputs_multi_subjects(self, data_rois, irois, all_outputs,
                                    targetAxes, ext, meta_data, output_dir):

        coutputs = {}
        output_fns = []

        roi_masks = [[(ds.roiMask != ds.backgroundLabel)
                      for ds in dr.data_subjects]
                     for dr in data_rois]
        #-> roi_masks[roi_id][subject]

        np_roi_masks = [[np.where(roi_subj_mask)
                         for roi_subj_mask in roi_subj_masks]
                        for roi_subj_masks in roi_masks]
        #-> np_roi_masks[roi_id][subject]

        for output_name, roi_outputs in all_outputs.iteritems():
            logger.info('Merge output %s ...', output_name)
            dest_c = {}
            try:
                if roi_outputs[0].has_axis('voxel'):
                    if not roi_outputs[0].has_axis('subject'):
                        raise Exception('Voxel-mapped output "%s" does not'
                                        'have a subject axis')
                    logger.debug('Merge as expansion ...')
                    dest_c_tmp = None
                    for isubj in roi_outputs[0].get_domain('subject'):
                        for i_roi, c in enumerate(roi_outputs):
                            m = np_roi_masks[i_roi][isubj]
                            dest_c_tmp = c.expand(roi_masks[i_roi][isubj],
                                                  'voxel', targetAxes,
                                                  dest=dest_c_tmp,
                                                  do_checks=False,
                                                  m=m)
                    dest_c[output_name] = dest_c_tmp
                else:
                    logger.debug('Merge as stack (%d elements)...', len(irois))
                    c_to_stack = [roi_outputs[i] for i in np.argsort(irois)]
                    dest_c[output_name] = stack_cuboids(c_to_stack,
                                                        domain=sorted(irois),
                                                        axis='ROI')
            except Exception, e:
                logger.error("Could not merge outputs for %s", output_name)
                logger.error("Exception was:")
                logger.error(e)

            for output_name, c in dest_c.iteritems():
                output_fn = op.join(output_dir, output_name + ext)
                output_fn = add_prefix(output_fn, self.outPrefix)
                output_fns.append(output_fn)
                logger.debug('Save output %s to %s', output_name, output_fn)
                try:
                    c.save(output_fn, meta_data=meta_data,
                           set_MRI_orientation=True)
                except Exception:
                    print 'Could not save output "%s", error stack was:' \
                        % output_name
                    exc_type, exc_value, exc_traceback = sys.exc_info()
                    traceback.print_exception(exc_type, exc_value,
                                              exc_traceback,
                                              limit=4, file=sys.stdout)
                coutputs[output_name] = c

        return coutputs, output_fns

    def outputResults(self, results, output_dir, filter='.\A',):
        """
        Return: a tuple (dictionary of outputs, output file names)
        """
        if output_dir is None:
            return {}, []
        # print results
        if not isinstance(results[0][0], (FmriData, FmriGroupData)):
            self.outputResults_back_compat(results, output_dir, filter)
            return {}, []

        logger.info('Building outputs from %d results ...', len(results))
        logger.debug('results :')
        logger.debug(results)

        # Handle analyses that crashed
        results = self.filter_crashed_results(results)

        if len(results) == 0:
            logger.error('No result to treat. Did everything crash ?')
            return {}, []

        # Merge all the analysis outputs
        target_shape = results[0][0].spatial_shape
        meta_data = results[0][0].meta_obj

        if len(target_shape) == 3:  # Volumic data:
            targetAxes = MRI3Daxes  # ['axial','coronal', 'sagittal']
            ext = '.nii'
        else:  # surfacic
            targetAxes = ['voxel']
            ext = '.gii'

        if self.gzip_outputs:
            ext += '.gz'

        if hasattr(results[0][1], 'getOutputs'):
            all_outputs = stack_trees([r[1].getOutputs() for r in results])
        else:
            all_outputs = stack_trees([r[1] for r in results])

        data_rois = [r[0] for r in results]
        irois = [d.get_roi_id() for d in data_rois]

        if isinstance(results[0][0], FmriData):
            return self.make_outputs_single_subject(data_rois, irois, all_outputs,
                                                    targetAxes, ext, meta_data,
                                                    output_dir)
        else:
            return self.make_outputs_multi_subjects(data_rois, irois, all_outputs,
                                                    targetAxes, ext, meta_data,
                                                    output_dir)

    def enable_draft_testing(self):
        raise NotImplementedError(
            'Enabling of draft testing is not implemented')

    def outputResults_back_compat(self, results, output_dir, filter='.\A',):

        logger.warning('Content of result.pck seems outdated, consider '
                       'running the analysis again to update it.')

        if output_dir is None:
            return

        logger.info('Building outputs ...')
        logger.debug('results :')
        logger.debug(results)

        to_pop = []
        for i, r in enumerate(results[:]):
            roi_id, result, report = r
            if report != 'ok':
                logger.error('-> Sampling crashed, roi %d!', roi_id)
                logger.error(report)
                to_pop.insert(0, i)

            elif result is None:
                logger.error('-> Sampling crashed (result is None), roi %d!',
                            roi_id)
                to_pop.insert(0, i)

        for i in to_pop:
            results.pop(i)

        if len(results) == 0:
            logger.warning('No more result to treat. Did everything crash ?')
            return

        def load_any(fns, load_func):
            for f in fns:
                try:
                    return load_func(f)
                except Exception:
                    pass
            return None

        r = load_any(['roiMask.tex', 'roi_mask.tex', 'jde_roi_mask.tex'],
                     read_texture)
        if r is None:
            r = load_any(['roi_mask.nii', 'jde_roi_mask.nii'], read_volume)

        if r is None:
            raise Exception('Can not find mask data file in current dir')

        all_rois_mask, meta_data = r

        target_shape = all_rois_mask.shape

        if len(target_shape) == 3:  # Volumic data:
            targetAxes = MRI3Daxes  # ['axial','coronal', 'sagittal']
            ext = '.nii'
        else:  # surfacic
            targetAxes = ['voxel']
            ext = '.gii'

        def genzip(gens):
            while True:
                # TODO: handle dict ...
                yield [g.next() for g in gens]

        logger.info('Get each ROI output ...')
        # print 'roi outputs:'
        if hasattr(results[0][1], 'getOutputs'):
            gen_outputs = [r[1].getOutputs() for r in results]
        else:
            gen_outputs = [r[1].iteritems() for r in results]
        irois = [r[0] for r in results]

        for roi_outputs in genzip(gen_outputs):
            output_name = roi_outputs[0][0]
            logger.info('Merge output %s ...', output_name)
            if roi_outputs[0][1].has_axis('voxel'):
                logger.info('Merge as expansion ...')
                dest_c = None
                for iroi, output in zip(irois, roi_outputs):
                    _, c = output
                    if output_name == 'ehrf':
                        print 'Before expansion:'
                        print c.descrip()

                        hrfs = c.data[0, :, :]
                        print 'ehrf for cond 0', hrfs.shape
                        for ih in xrange(hrfs.shape[1]):
                            print np.array2string(hrfs[:, ih])

                        # print np.array2string(hrfs, precision=2)

                    roi_mask = (all_rois_mask == iroi)
                    dest_c = c.expand(roi_mask, 'voxel', targetAxes,
                                      dest=dest_c)
                    if output_name == 'ehrf':
                        print 'After expansion:'
                        print dest_c.descrip()
                        m = np.where(roi_mask)
                        hrfs = dest_c.data[0, :, m[0], m[1], m[2]]
                        print 'ehrf for cond 0', hrfs.shape
                        for ih in xrange(hrfs.shape[0]):
                            print np.array2string(hrfs[ih, :])
            else:
                logger.info('Merge as stack (%d elements)...', len(irois))
                c_to_stack = [roi_outputs[i][1] for i in np.argsort(irois)]
                dest_c = stack_cuboids(c_to_stack, domain=sorted(irois),
                                       axis='ROI')

            output_fn = op.join(output_dir, output_name + ext)
            logger.info('Save output %s to %s', output_name, output_fn)
            try:
                dest_c.save(output_fn, meta_data=meta_data,
                            set_MRI_orientation=True)
            except Exception:
                print 'Could not save output "%s", error stack was:' \
                    % output_name
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_exception(exc_type, exc_value, exc_traceback,
                                          limit=4, file=sys.stdout)

    def joinOutputs(self, cuboids, roiIds, mappers):

        mapper0 = mappers[roiIds[0]]
        if mapper0.isExpendable(cuboids[0]):
            logger.info('Expanding ...')
            expandedxndarray = mapper0.expandxndarray(cuboids[0])
            for c, roi in zip(cuboids[1:], roiIds[1:]):
                mappers[roi].expandxndarray(c, dest=expandedxndarray)
            return expandedxndarray
        else:
            return stack_cuboids(cuboids, 'ROI', roiIds)

    def clean_output_files(self, output_dir):
        # Clean all formatted outputs
        if self.outFile is not None and output_dir is not None:
            out_file = op.join(output_dir, self.outFile)
            if op.exists(self.outFile):
                allOuts = xmlio.from_xml(open(out_file).read())
                for c in allOuts.itervalues():
                    c.cleanFiles()
                os.remove(out_file)



# -*- coding: utf-8 -*-

import os
import os.path as op
import re
import tempfile
import logging

from time import strftime, localtime, time

import numpy as np
from numpy.random import rand, randint, permutation
from nipy.labs.spatial_models.parcel_io import fixed_parcellation

try:
    from nipy.algorithms.clustering.clustering import voronoi
except ImportError:
    from nipy.algorithms.clustering.utils import voronoi

import pyhrf

from pyhrf.tools._io import read_volume, write_volume, read_texture, write_texture
from pyhrf.tools import format_duration, peelVolume3D
from pyhrf.graph import parcels_to_graphs, kerMask3D_6n, connected_components
from pyhrf.ndarray import expand_array_in_mask
from pyhrf.glm import glm_nipy


logger = logging.getLogger(__name__)

old_to_new_log_dict = {
    0: logging.WARNING,
    1: logging.INFO,
    2: logging.INFO,
    3: logging.INFO,
    4: logging.INFO,
    5: logging.DEBUG,
    6: logging.DEBUG,
}

def round_nb_parcels(n):
    if n >= 100:
        return int(np.round(n / 100) * 100)
    elif n >= 10:
        return int(np.round(n / 10) * 10)
    else:
        return int(np.round(n))


def parcellation_report(d):
    s = ''
    d = d.astype(int)
    s += 'Parcellation description:\n'
    roiIds = np.unique(d)
    m = d.min()
    id_shift = int(min(m, 0))
    counts = np.bincount(d.astype(int).flatten() - id_shift)
    if len(roiIds) == 1:
        s += '* One ROI, label=%d, size=%d' % (roiIds[0] + id_shift,
                                               counts[roiIds[0] + id_shift])
        return s

    if counts[0] > 0:
        avg_size = counts[counts != 0][1:].mean()
    else:
        avg_size = counts[counts != 0].mean()
    s += '* Nb voxels inside mask: %d\n' % (d > d.min()).sum()
    s += '* %d parcels, min=%d, max=%d, avg parcel size=%d\n'  \
        % (len(roiIds), m, d.max(), avg_size)
    s += '* Size of background: %s\n' % counts[0]
    counts_tmp = counts.astype(float)
    counts_tmp[counts == 0] = np.inf
    s += '* Id of smallest parcel: %d (size=%d)\n' \
        % (np.argmin(counts_tmp) + id_shift, int(counts_tmp.min()))
    s += '* Nb of parcels with size < 10: %d\n' % (counts_tmp < 10).sum()
    if 0:
        print m - id_shift + 1
        print zip(range(len(counts)), counts)
        print counts[1:].max()
        print sorted(counts)
    argmax = np.argmax(counts[1:]) + 1
    s += '* Id of biggest parcel: %d (size=%d)\n' \
        % (argmax + id_shift, counts[argmax])
    assert (d == (argmax + id_shift)).sum() == counts[argmax]
    s += '* Nb of parcels with size is > 400: %d\n' \
        % (counts[m - id_shift + 1:] > 400).sum()
    s += '* Nb of parcels with size is > 1000: %d\n' \
        % (counts[m - id_shift + 1:] > 1000).sum()

    return s


def parcellation_ward_spatial(func_data, n_clusters, graph=None):
    """Make parcellation based upon ward hierarchical clustering from scikit-learn

    Parameters
    ----------
    func_data: array of shape (nb_positions, dim_feature_1, [dim_feature2, ...])
        functional data:
    n_clusters
        chosen number of clusters to create
    graph
        adjacency list defining neighbours. if None, no connectivity defined: clustering is spatially independent

    Returns
    -------
    parcellation labels
    """
    try:
        # sklearn version < 0.17
        from sklearn.cluster import Ward as AgglomerativeClustering
    except ImportError:
        from sklearn.cluster import AgglomerativeClustering

    from pyhrf.graph import graph_to_sparse_matrix, sub_graph
    if graph is not None:
        labels = np.zeros(len(graph), dtype=np.int32)
        ccs = connected_components(graph)
        n_tot = len(graph) * 1.
        ncs = np.zeros(len(ccs), dtype=np.int32)
        for icc, cc in enumerate(ccs):
            nc = int(np.round(n_clusters * len(cc) / n_tot))
            if nc == 0:
                nc = 1
            ncs[icc] = nc

        max_nc = np.argmax(ncs)
        ncs[max_nc] = n_clusters - sum(ncs[0:max_nc]) - sum(ncs[max_nc + 1:])

        assert sum(ncs) == n_clusters
        assert (ncs > 0).all()

        logger.info('Found %d connected components (CC) of sizes: %s', len(ccs),
                    ' ,'.join([str(len(cc)) for cc in ccs]))
        logger.info('Nb of clusters to search in each CC: %s',
                    ' ,'.join(map(str, ncs)))

        for nc, cc in zip(ncs, ccs):
            if len(cc) < 2:
                continue

            if len(cc) < len(graph):
                cc_graph, _ = sub_graph(graph, cc)
            else:
                cc_graph = graph

            cc_connectivity = graph_to_sparse_matrix(cc_graph)

            cc_data = func_data[cc]
            logger.info('Launch spatial Ward (nclusters=%d)  on data of shape %s',
                        nc, str(cc_data.shape))
            ward_object = AgglomerativeClustering(
                n_clusters=nc, connectivity=cc_connectivity
            ).fit(cc_data)
            labels[cc] += ward_object.labels_ + 1 + labels.max()
    else:
        ward_object = AgglomerativeClustering(
            n_clusters=n_clusters
        ).fit(func_data)  # connectivity=None
        labels = ward_object.labels_ + 1

    return labels


def make_parcellation_from_files(betaFiles, maskFile, outFile, nparcels,
                                 method, dry=False, spatial_weight=10.):

    if not op.exists(maskFile):
        print 'Error, file %s not found' % maskFile
        return

    betaFiles = sorted(betaFiles)
    for b in betaFiles:
        if not op.exists(b):
            raise Exception('Error, file %s not found' % b)

    logger.info('Mask image: %s', op.basename(maskFile))
    logger.info('Betas: %s ... %s', op.basename(betaFiles[0]),
                op.basename(betaFiles[-1]))
    logger.info("Method: %s, nb parcels: %d", method, nparcels)
    logger.info('Spatial weight: %f', spatial_weight)

    if not dry:
        logger.info('Running parcellation ... ')
        logger.info('Start date is: %s', strftime('%c', localtime()))
        t0 = time()
        v = logger.getEffectiveLevel() <= logging.INFO
        lpa = fixed_parcellation(maskFile, betaFiles, nparcels, nn=6,
                                 method=method, fullpath=outFile, verbose=v,
                                 mu=spatial_weight)

        from pyhrf.ndarray import xndarray
        c = xndarray.load(outFile)
        if c.min() == -1:
            c.data += 1

        for i in np.unique(c.data):
            # remove parcel with size < 2:
            if i != 0 and len(c.data == 1) < 2:
                c.data[np.where(c.data == i)] = 0

        c.save(outFile)

        logger.info(
            'Parcellation complete, took %s', format_duration(time() - t0))
        return lpa
    else:
        logger.info('Dry run.')


from nipy.labs.spatial_models.discrete_domain import domain_from_mesh
from nipy.algorithms.graph.field import field_from_coo_matrix_and_data
from nipy.labs.spatial_models.mroi import SubDomains


def make_parcellation_surf_from_files(beta_files, mesh_file, parcellation_file,
                                      nbparcel, method, mu=10., verbose=0):

    if method not in ['ward', 'gkm', 'ward_and_gkm', 'kmeans']:
        raise ValueError('unknown method')

    # step 1: load the data ----------------------------
    # 1.1 the domain
    logger.info('domain from mesh: %s', mesh_file)
    domain = domain_from_mesh(mesh_file)

    coord = domain.coord

    # 1.3 read the functional data
    beta = np.array([read_texture(b)[0] for b in beta_files]).T

    logger.info('beta: %s', str(beta.shape))
    logger.info('mu * coord / np.std(coord): %s',
                (mu * coord / np.std(coord)).shape)
    feature = np.hstack((beta, mu * coord / np.std(coord)))

    if method is not 'kmeans':
        g = field_from_coo_matrix_and_data(domain.topology, feature)

    if method == 'kmeans':
        _, u, _ = kmeans(feature, nbparcel)

    if method == 'ward':
        u, _ = g.ward(nbparcel)

    if method == 'gkm':
        seeds = np.argsort(np.random.rand(g.V))[:nbparcel]
        _, u, _ = g.geodesic_kmeans(seeds)

    if method == 'ward_and_gkm':
        w, _ = g.ward(nbparcel)
        _, u, _ = g.geodesic_kmeans(label=w)

    lpa = SubDomains(domain, u, 'parcellation')

    if verbose:
        var_beta = np.array(
            [np.var(beta[lpa.label == k], 0).sum() for k in range(lpa.k)])
        var_coord = np.array(
            [np.var(coord[lpa.label == k], 0).sum() for k in range(lpa.k)])
        size = lpa.get_size()
        vf = np.dot(var_beta, size) / size.sum()
        va = np.dot(var_coord, size) / size.sum()
        print nbparcel, "functional variance", vf, "anatomical variance", va

    # step3:  write the resulting label image
    if parcellation_file is not None:
        label_image = parcellation_file
    else:
        label_image = None

    if label_image is not None:
        write_texture(u.astype(np.int32), label_image)
        if verbose:
            print "Wrote the parcellation images as %s" % label_image

    return u, label_image


#############################################################
## Balanced partitioning using balloon-like a(ge)nt system ##
#############################################################


def random_pick(a):
    return a[randint(0, a.size)]


def init_edge_data(g, init_value=0):
    return [dict([(i, init_value) for i in nl]) for nl in g]

# TODO: replace the following class by logging module


class Talker:

    def __init__(self, talker_string_id, verbosity=0):
        self.verbosity = verbosity
        if len(talker_string_id) > 0:
            self.t_id = '%s ||' % talker_string_id
        else:
            self.t_id = ''

    def verbose(self, level, msg):
        logger.log(old_to_new_log_dict[level], msg)

    def verbose_array(self, level, array):
        if self.verbosity >= level:
            stars = '*' * level
            if array.ndim > 1:
                for l in array:
                    print stars, self.t_id, l
            else:
                print stars, self.t_id, array


def Visit_graph_noeud(noeud, graphe, Visited=None):
    if Visited is None:
        Visited = np.zeros(len(graphe), dtype=int)

    Visited[noeud] = 1
    for voisin in graphe[noeud]:
        if not Visited[voisin]:
            Visit_graph_noeud(voisin, graphe, Visited)

    return Visited


class World(Talker):

    def __init__(self, graph, nb_ants, greed=.05,
                 time_min=100, time_max=None, tolerance=1, verbosity=0,
                 stop_when_all_controlled=False):

        Talker.__init__(self, '', verbosity)
        self.tolerance = tolerance
        self.time_max = time_max
        self.time_min = time_min
        self.nb_ants = nb_ants
        self.graph = graph
        self.nb_sites = len(self.graph)
        self.labels = np.zeros(self.nb_sites, dtype=np.int32) - 1
        self.path_marks = init_edge_data(self.graph)
        self.site_marks = np.zeros(self.nb_sites, dtype=np.int32)
        self.pressures = init_edge_data(self.graph)

        self.ants = [Ant(i, greed, self.graph, self.labels, self.path_marks,
                         self.site_marks, self.pressures, self, verbosity)
                     for i in xrange(self.nb_ants)]

        self.uncontrolled = self.nb_sites - self.nb_ants

        self.tmark = time()
        self.prev_uncontrolled = self.nb_sites
        self.total_time = .0

        self.stop_at_full_control = stop_when_all_controlled

    def balanced(self):
        if self.uncontrolled > 0:
            return False

        if self.time < self.time_min:
            self.verbose(6, 'time_min not reached')
            return False

        return np.allclose([a.area_size for a in self.ants],
                           self.nb_sites / self.nb_ants, atol=self.tolerance)

    def force_end(self):
        if self.stop_at_full_control and self.uncontrolled == 0:
            return True

        if self.time_max is None:
            return False
        else:
            return self.time >= self.time_max

    def site_taken(self, site):
        ant_mark = self.labels[site]
        if ant_mark != -1:
            self.ants[ant_mark].territory_size -= 1

    def resolve(self):

        t0 = time()
        self.time = 0
        while not self.balanced() and not self.force_end():
            for a_id in permutation(self.nb_ants):
                if self.ants[a_id].action(self.time):
                    continue
            self.time += 1
            if self.time % 100000 == 0:
                self.verbose(1, 'Time: %d' % self.time)

                self.verbose(1, 'Nb of uncontrolled sites: %d' %
                             self.uncontrolled)
                sizes = [str(a.area_size) for a in self.ants]
                self.verbose(1, 'Territories: %s' % ','.join(sizes))
                if self.uncontrolled > 0:
                    delta_t = time() - self.tmark
                    duration = format_duration(delta_t)
                    delta_site = self.prev_uncontrolled - self.uncontrolled
                    self.verbose(1, '%d sites taken in %s'
                                 % (delta_site, duration))
                    self.total_time += delta_t
                    speed = (self.nb_sites - self.uncontrolled) / \
                        self.total_time
                    expected_duration = self.uncontrolled / speed
                    self.verbose(1, '%s expected for complete control'
                                 % (format_duration(expected_duration)))

                    self.prev_uncontrolled = self.uncontrolled
                    self.tmark = time()

        self.verbose(1, 'Balloon partitioning done, took %s, %d iterations'
                     % (format_duration(time() - t0), self.time))

    def get_final_labels(self):
        return self.labels


class Ant(Talker):

    def __init__(self, a_id, greed, graph, labels, path_marks, site_marks,
                 pressures, world, verbosity=0):
        Talker.__init__(self, 'ant %d' % a_id, verbosity)
        self.world = world
        self.search_level = 0
        self.area_size = 1
        self.labels = labels
        self.path_marks = path_marks
        self.site_marks = site_marks
        self.pressures = pressures
        self.graph = graph
        self.a_id = a_id
        self.greed = greed
        self.verbosity = verbosity

        # pick init position
        if self.labels[0] == -1:
            start_pos = 0
        elif self.labels[-1] == -1:
            start_pos = len(self.labels) - 1
        elif self.labels[len(self.labels) / 2] == -1:
            start_pos = len(self.labels) / 2
        else:
            unmarked = np.where(self.labels == -1)
            picked = randint(len(unmarked[0]))
            start_pos = unmarked[0][picked]

        self.labels[start_pos] = a_id

        self.u = np.where(self.labels == self.a_id)[0][0]  # current position

    def to_conquer(self):
        for v in self.graph[self.u]:
            if self.labels[v] != self.a_id and \
                    self.path_marks[self.u][v] < self.search_level:
                yield v

    def to_patrol(self):
        for v in self.graph[self.u]:
            if self.labels[v] == self.a_id and \
                    self.path_marks[self.u][v] < self.search_level:
                yield v

    def fix_explosion(self):
        if self.world.uncontrolled < 1:
            for v in self.graph[self.u]:
                if self.labels[v] != self.a_id and self.labels[v] != -1:
                    orphan = True
                    for w in self.graph[v]:
                        if self.labels[w] != -1 and \
                                self.labels[w] == self.labels[v]:
                            orphan = False
                            break
                    if orphan:
                        self.labels[v] = -1
                        self.world.uncontrolled += 1
                        return

    def action(self, time):

        if self.labels[self.u] != self.a_id:
            owned_neighbours = self.labels[self.graph[self.u]] == self.a_id
            if owned_neighbours.sum() > 0:
                self.u = random_pick(self.graph[self.u][owned_neighbours])
                return False

        # Try to conquer not owned nearby sites:
        for v in self.to_conquer():
            # activate exploration of path:
            self.path_marks[self.u][v] = time
            if self.pressures[self.u][v] > 0:
                if rand() < self.greed:
                    for w in self.graph[v]:
                        self.path_marks[v][w] = time + 1
                        self.pressures[v][w] = 0
                        self.pressures[w][v] = 0
                    self.path_marks[v][self.u] = time
                    self.site_marks[v] = time
                    if self.labels[v] == -1:
                        self.world.uncontrolled -= 1
                    self.labels[v] = self.a_id
                    self.u = v
                    self.fix_explosion()
                    return False
                else:
                    pass
            else:
                if self.pressures[self.u][v] + self.area_size == -1:
                    self.pressures[self.u][v] = 0
                    self.pressures[v][self.u] = 0
                else:
                    self.pressures[self.u][v] = self.area_size
                    self.pressures[v][self.u] = -self.area_size

        # Partrol in nearby owned site
        for v in self.to_patrol():
            self.path_marks[self.u][v] = time
            if self.site_marks[v] < self.search_level:
                self.area_size = (time - self.site_marks[v] + 1) / 2
                self.path_marks[v][self.u] = time
                self.site_marks[v] = time
                self.u = v
                return False

        for v in self.graph[self.u]:
            if self.labels[v] == self.a_id and \
                    self.site_marks[self.u] > self.search_level and \
                    self.path_marks[self.u][v] == self.site_marks[self.u]:
                self.u = v
                return False

        # Agent didn't do anything
        self.search_level = time
        self.site_marks[self.u] = time
        for w in self.graph[self.u]:
            self.path_marks[self.u][w] -= 1

        return True


import pyhrf.graph as mg


def parcellate_balanced_vol(mask, nb_parcels):
    """
    Performs a balanced partitioning on the input mask using a balloon patroling
     algorithm [Eurol 2009]. Values with 0 are discarded position in the mask.

    Args:
        - mask (numpy.ndarray): binary 3D array of valid position to parcellate
        - nb_parcels (int): the required number of parcels

    Return:
        - the parcellation (numpy.ndarray): a 3D array of integers
    """

    parcellation = np.zeros(mask.shape, dtype=int)
    nvox = (mask != 0).sum()

    # Iterate over connected components in the input mask
    for cc_mask in mg.split_mask_into_cc_iter(mask != 0):
        logger.info('Treating a connected component (CC) of %d positions',
                    cc_mask.sum())
        g = mg.graph_from_lattice(cc_mask)
        size_cc = cc_mask.sum()
        # compute the required number of parcels within the current CC:
        cc_np = max(int(np.round(nb_parcels * size_cc / (nvox * 1.))), 1)

        logger.info('Split (CC) into %d parcels', cc_np)

        cc_labels = np.ones(cc_mask.sum(), dtype=int)
        if cc_np > 1:
            split_parcel(cc_labels, {1: g}, 1, cc_np, inplace=True,
                         verbosity=2, balance_tolerance='draft')
        else:  # only one parcel expected, CC must be too small to be splited
            cc_labels[:] = 1
        logger.info('Split done!')

        # accumulate parcellation result
        maxp = parcellation.max()
        parcellation += expand_array_in_mask(cc_labels + maxp, cc_mask > 0)

    return parcellation


def split_parcel(labels, graphs, id_parcel, n_parcels, inplace=False,
                 verbosity=0, balance_tolerance='exact'):
    """
    balance_tolerance : exact or draft
    """
    g = graphs[id_parcel]
    nb_cc = len(connected_components(g))
    if nb_cc > 1:
        raise Exception("The input graph is not connex "
                        "(%d connected components found)" % nb_cc)
    world = World(g, n_parcels, verbosity=verbosity,
                  stop_when_all_controlled=(balance_tolerance == 'draft'))
    world.resolve()

    fl = world.get_final_labels()
    fl[fl > 0] += labels.max()
    fl[fl == 0] = id_parcel

    if not inplace:
        new_labels = labels.copy()
        input_labels = labels
    else:
        input_labels = labels.copy()
        new_labels = labels

    if 0:
        print 'new_labels:', new_labels.shape
        print 'input_labels:', input_labels.shape, new_labels.dtype
        print 'id_parcel:', id_parcel
        print 'new_labels[np.where(input_labels==id_parcel)]:', \
            new_labels[np.where(input_labels == id_parcel)].shape
        print np.unique(new_labels[np.where(input_labels == id_parcel)])
        print 'fl:', fl.shape, fl.dtype
        print np.unique(fl)

    new_labels[np.where(input_labels == id_parcel)] = fl
    return new_labels


def split_big_parcels(parcel_file, output_file, max_size=400):
    print 'split_big_parcels ...'
    roiMask, roiHeader = read_volume(parcel_file)
    roiIds = np.unique(roiMask)
    background = roiIds.min()
    labels = roiMask[np.where(roiMask > background)].astype(int)
    if (np.bincount(labels) <= max_size).all():
        logger.info('no parcel to split')
        return

    graphs = parcels_to_graphs(roiMask, kerMask3D_6n)
    for roiId in roiIds:
        if roiId != background:
            roi_size = (roiMask == roiId).sum()
            if roi_size > max_size:
                print 'roi %d, size = %d' % (roiId, roi_size)
                nparcels = int(np.ceil(roi_size * 1. / max_size))
                print 'split into %d parcels ...' % (nparcels)
                split_parcel(labels, graphs, roiId, nparcels, inplace=True,
                             verbosity=1)

    final_roi_mask = np.zeros_like(roiMask)
    final_roi_mask[np.where(roiMask > background)] = labels
    assert (np.bincount(labels) <= max_size).all()
    write_volume(final_roi_mask, output_file, roiHeader)


def parcellate_voronoi_vol(mask, nb_parcels, seeds=None):
    """
    Produce a parcellation from a Voronoi diagram built on random seeds.
    The number of seeds is equal to the nb of parcels.
    Seed are randomly placed within the mask, expect on edge positions

    Args:
        - mask (numpy.ndarray): binary 3D array of valid position to parcellate
        - nb_parcels (int): the required number of parcels
        - seeds: TODO

    Return:
        - the parcellation (numpy.ndarray): a 3D array of integers
        -
    """
    parcellation = np.zeros(mask.shape, dtype=int)
    nvox = (mask != 0).sum()
    for cc_mask in mg.split_mask_into_cc_iter(mask != 0):
        # compute the required number of parcels within the current CC:
        size_cc = cc_mask.sum()
        cc_np = max(int(np.round(nb_parcels * size_cc / (nvox * 1.))), 1)
        logger.info('Treating a connected component (CC) of %d positions',
                    cc_mask.sum())
        if cc_mask.sum() < 6:
            continue
        if seeds is None:
            # perform voronoi on random seeds

            eroded_mask = peelVolume3D(cc_mask)
            eroded_mask_size = eroded_mask.sum()
            if eroded_mask_size < nb_parcels:  # do no erode, mask too small
                eroded_mask_size = nvox
                eroded_mask = mask.copy()
            cc_seeds = np.random.randint(0, eroded_mask_size, cc_np)
            mask_for_seed = np.zeros(eroded_mask_size, dtype=int)
            mask_for_seed[cc_seeds] = 1
            mask_for_seed = expand_array_in_mask(mask_for_seed, eroded_mask)
        else:
            mask_for_seed = seeds * cc_mask

        logger.info('Nb of seeds in current CC: %d', mask_for_seed.sum())
        cc_parcellation = voronoi(np.vstack(np.where(cc_mask)).T,
                                  np.vstack(np.where(mask_for_seed)).T) + 1
        logger.info(
            'CC parcellation labels: %s', str(np.unique(cc_parcellation)))
        maxp = parcellation.max()
        parcellation += expand_array_in_mask(cc_parcellation + maxp, cc_mask)
        logger.info('Current parcellation labels: %s',
                    str(np.unique(parcellation)))
    logger.info('voronoi parcellation: %s, %s', str(parcellation.shape),
                str(parcellation.dtype))

    return parcellation


def parcellation_for_jde(fmri_data, avg_parcel_size=250, output_dir=None,
                         method='gkm', glm_drift='Cosine', glm_hfcut=128):
    """
    method: gkm, ward, ward_and_gkm
    """

    if output_dir is None:
        output_dir = tempfile.mkdtemp(prefix='pyhrf_JDE_parcellation_GLM',
                                      dir=pyhrf.cfg['global']['tmp_path'])
    glm_output_dir = op.join(output_dir, 'GLM_for_parcellation')
    if not op.exists(glm_output_dir):
        os.makedirs(glm_output_dir)

    logger.info('GLM for parcellation')

    g, dm, cons = glm_nipy(fmri_data, drift_model=glm_drift, hfcut=glm_hfcut)

    pval_files = []
    if cons is not None:
        func_data = [('con_pval_%s' % cname, con.pvalue())
                     for cname, con in cons.iteritems()]
    else:
        reg_cst_drift = re.compile(".*constant.*|.*drift.*")
        func_data = [('beta_%s' % reg_name, g.beta[ir])
                     for ir, reg_name in enumerate(dm.names)
                     if not reg_cst_drift.match(reg_name)]

    for name, data in func_data:
        val_vol = expand_array_in_mask(data, fmri_data.roiMask > 0)
        val_fn = op.join(glm_output_dir, '%s.nii' % name)
        write_volume(val_vol, val_fn, fmri_data.meta_obj)
        pval_files.append(val_fn)

    mask_file = op.join(glm_output_dir, 'mask.nii')
    write_volume(fmri_data.roiMask > 0, mask_file, fmri_data.meta_obj)

    nvox = fmri_data.get_nb_vox_in_mask()
    nparcels = round_nb_parcels(nvox * 1. / avg_parcel_size)

    logger.info('Parcellation from GLM outputs, method: %s, nb parcels: %d',
                method, nparcels)

    if fmri_data.data_type == 'volume':
        parcellation_file = op.join(output_dir, 'parcellation_%s_np%d.nii'
                                    % (method, nparcels))

        make_parcellation_from_files(pval_files, mask_file, parcellation_file,
                                     nparcels, method)
        parcellation, _ = read_volume(parcellation_file)
    else:
        mesh_file = fmri_data.data_files[-1]
        parcellation_file = op.join(output_dir, 'parcellation_%s_np%d.gii'
                                    % (method, nparcels))
        make_parcellation_surf_from_files(pval_files, mesh_file,
                                          parcellation_file, nparcels, method,
                                          verbose=1)
        parcellation, _ = read_texture(parcellation_file)

    logger.info(parcellation_report(parcellation))

    return parcellation, parcellation_file


def make_parcellation_cubed_blobs_from_file(parcellation_file, output_path,
                                            roi_ids=None, bg_parcel=0,
                                            skip_existing=False):

    p, mp = read_volume(parcellation_file)
    p = p.astype(np.int32)
    if bg_parcel == 0 and p.min() == -1:
        p += 1  # set background to 0

    if roi_ids is None:
        roi_ids = np.unique(p)

    logger.info('%d rois to extract', (len(roi_ids) - 1))

    tmp_dir = pyhrf.get_tmp_path('blob_parcellation')
    tmp_parcel_mask_file = op.join(tmp_dir, 'parcel_for_blob.nii')

    out_files = []
    for roi_id in roi_ids:
        if roi_id != bg_parcel:  # discard background
            output_blob_file = op.join(output_path, 'parcel_%d_cubed_blob.arg'
                                       % roi_id)
            out_files.append(output_blob_file)
            if skip_existing and os.path.exists(output_blob_file):
                continue
            parcel_mask = (p == roi_id).astype(np.int32)
            write_volume(parcel_mask, tmp_parcel_mask_file, mp)
            logger.info('Extract ROI %d -> %s', roi_id, output_blob_file)
            cmd = 'AimsGraphConvert -i %s -o %s --bucket' \
                % (tmp_parcel_mask_file, output_blob_file)
            logger.info('Cmd: %s', cmd)
            os.system(cmd)
    if op.exists(tmp_parcel_mask_file):
        os.remove(tmp_parcel_mask_file)

    return out_files


from pyhrf.cparcellation import compute_intersection_matrix


def parcellation_dist(p1, p2, mask=None):
    """
    Compute the distance between the two parcellation p1 and p2 as the minimum
    number of positions to remove in order to obtain equal partitions.
    "mask" may be a binary mask to limit the distance computation to some
    specific positions.
    Important convention: parcel label 0 is treated as background and
    corresponding positions are discarded if no mask is provided.

    Return:
        (distance value, parcellation overlap)
    """
    assert np.issubdtype(p1.dtype, np.int)
    assert np.issubdtype(p2.dtype, np.int)

    assert p1.shape == p2.shape

    from munkres import Munkres
    if mask is None:
        mask = (p1 != 0)

    m = np.where(mask)
    logger.debug('Nb pos inside mask: %d', len(m[0]))

    fp1 = p1[m].astype(np.int32)
    fp2 = p2[m].astype(np.int32)

    # allocate cost matrix, assume that region labels are contiguous
    # ie all labels in [1, label_max] are represented
    cost_matrix = np.zeros((fp1.max() + 1, fp2.max() + 1), dtype=np.int32)

    logger.debug('Cost matrix : %s', str(cost_matrix.shape))
    compute_intersection_matrix(fp1, fp2, cost_matrix)

    # discard 0-labelled parcels (background)
    cost_matrix = cost_matrix[1:, 1:]

    # solve the assignement problem:
    indexes = np.array(Munkres().compute((cost_matrix * -1).tolist()))

    if 0:
        print 'assignement indexes:'
        print indexes
        print '->'
        print (indexes[:, 0], indexes[:, 1])
        print cost_matrix[(indexes[:, 0], indexes[:, 1])]

    assignement = cost_matrix[(indexes[:, 0], indexes[:, 1])].sum()

    to_keep = np.zeros_like(fp1)
    for s1, s2 in indexes:
        to_keep += np.bitwise_and(fp1 == (s1 + 1), fp2 == (s2 + 1))

    return fp1.size - assignement, expand_array_in_mask(to_keep, mask)



# -*- coding: utf-8 -*-


import pyhrf

savePlots = False
plotSaveDir = './'
figext = 'svg' #'png'
useCache = False
cacheDir = '/tmp'

def figfn(fn):
    """ Append the figure file extension to fn
    """
    print 'figext:',figext
    return fn + '.' + figext

def is_tmp_file(fn):
    return fn.startswith(pyhrf.tmpPrefix)

def clean_cache():
    tmpFiles = filter(is_tmp_file, os.listdir(cacheDir))
    for f in tmpFiles:
        os.remove(os.path.join(cacheDir, f))


    



# -*- coding: utf-8 -*-

import unittest
import pyhrf
import numpy as np
import os.path as op
import shutil

from pyhrf.ndarray import MRI3Daxes
import pyhrf.boldsynth.scenarios as simu
from pyhrf.rfir import rfir

class RFIRTest(unittest.TestCase):
    """
    Test the Regularized FIR (RFIR)-based methods implemented in pyhrf.rfir
    """

    def setUp(self):
        self.tmp_dir = pyhrf.get_tmp_path()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_rfir_on_small_simulation(self):
        """ Check if pyhrf.rfir runs properly and that returned outputs
        contains the expected items """
        fdata = simu.create_small_bold_simulation()
        outputs = rfir(fdata, nb_its_max=2)

        assert isinstance(outputs, dict)
        for k in ["fir", "fir_error", "fit", "drift"]:
            assert outputs.has_key(k)


    def test_results_small_simulation(self):
        """

        TODO: move to validation
        """
        #output_dir = pyhrf.get_tmp_path()
        output_dir = './'

        fdata = simu.create_small_bold_simulation(output_dir=output_dir)

        # pyhrf.verbose.set_verbosity(0)
        pyhrf.logger.setLevel(logging.WARNING)
        outputs = rfir(fdata, nb_its_max=500, nb_its_min=100)

        m = fdata.roiMask
        for k,o in outputs.iteritems():
            output_fn = op.join(output_dir, k + '.nii')
            o.expand(m, "voxel", target_axes=MRI3Daxes).save(output_fn)



# -*- coding: utf-8 -*-

import os
import os.path as op
import sys
import string
import cPickle
import hashlib
import gzip
import datetime
import inspect
import re
import logging

from itertools import izip
from time import time
from collections import defaultdict

import numpy as np
import scipy.linalg
import scipy.signal

import pyhrf

try:
    from itertools import product as iproduct
except ImportError:
    def iproduct(*args, **kwds):
        # product('ABCD', 'xy') --> Ax Ay Bx By Cx Cy Dx Dy
        # product(range(2), repeat=3) --> 000 001 010 011 100 101 110 111
        pools = map(tuple, args) * kwds.get('repeat', 1)
        result = [[]]
        for pool in pools:
            result = [x + [y] for x in result for y in pool]
            for prod in result:
                yield tuple(prod)


logger = logging.getLogger(__name__)


class PickleableStaticMethod(object):
    def __init__(self, fn, cls=None):
        self.cls = cls
        self.fn = fn
        self.__name__ = fn.__name__

    def __call__(self, *args, **kwargs):
        if self.cls is None:
            return self.fn(*args, **kwargs)
        else:
            return self.fn(self.cls, *args, **kwargs)

    def __get__(self, obj, cls):
        return PickleableStaticMethod(self.fn, cls)

    def __getstate__(self):
        return (self.cls, self.fn.__name__)

    def __setstate__(self, state):
        self.cls, name = state
        self.fn = getattr(self.cls, name).fn


def is_importable(module_name, func_name=None):
    """ Return True if given *module_name* (str) is importable """
    try:
        __import__(module_name)
    except ImportError:
        return False
    else:
        if func_name is None:
            return True
        else:
            return func_name in dir(module_name)


def report_arrays_in_obj(o):
    for a in dir(o):
        attr = eval('o.%s' % a)
        if isinstance(attr, np.ndarray):
            print a, '->', attr.shape, attr.dtype, '[id=%s]' % str(id(attr))
        elif (isinstance(attr, list) or isinstance(attr, tuple)) and len(attr) > 0 and \
                isinstance(attr[0], np.ndarray):
            print a, '-> list of %d arrays (first array: %s, %s), [id=%s]' \
                % (len(attr), str(attr[0].shape), str(attr[0].dtype), str(id(attr)))
        elif isinstance(attr, dict) \
                and len(attr) > 0 and isinstance(attr[attr.keys()[0]], np.ndarray):
            print a, 'is a dict of arrays comprising:'
            for k, v in attr.iteritems():
                print k, '->'
                report_arrays_in_obj(v)
            print 'end of listing dict "%s"' % a


def has_ext(fn, ext):
    if fn.endswith('gz'):
        return fn.split('.')[-2] == ext
    else:
        return fn.split('.')[-1] == ext


def replace_ext(fn, ext):
    if fn.endswith('gz'):
        return '.'.join(fn.split('.')[:-2] + [ext, 'gz'])
    else:
        return '.'.join(fn.split('.')[:-1] + [ext])


def add_suffix(fn, suffix):
    """ Add a suffix before file extension.

    >>> add_suffix('./my_file.txt', '_my_suffix')
    './my_file_my_suffix.txt'
    """
    if suffix is None:
        return fn
    sfn = op.splitext(fn)
    if sfn[1] == '.gz':
        sfn = op.splitext(fn[:-3])
        sfn = (sfn[0], sfn[1] + '.gz')
    return sfn[0] + suffix + sfn[1]


def add_prefix(fn, prefix):
    """ Add a prefix at the beginning of a file name.

    >>> add_prefix('./my_file.txt', 'my_prefix_')
    './my_prefix_my_file.txt'
    """

    if prefix is None:
        return fn
    sfn = op.split(fn)
    if sfn[1] == '.gz':
        sfn = op.splitext(fn[:-3])
        sfn = (sfn[0], sfn[1] + '.gz')

    return op.join(sfn[0], prefix + sfn[1])


def assert_path_not_in_src(p):
    p = op.realpath(p)
    src_path = op.realpath(op.join(op.dirname(pyhrf.__file__), '../../'))
    for root, _, _ in os.walk(src_path):
        if root == p:
            raise Exception('Directory %s is in source path' % p)


def assert_file_exists(fn):
    if not op.exists(fn):
        raise Exception('File %s does not exists' % fn)


def non_existent_canditate(f, start_idx=1):
    yield f
    i = start_idx
    while True:
        yield add_suffix(f, '_%d' % i)
        i += 1


def non_existent_file(f):
    for f in non_existent_canditate(f):
        if not op.exists(f):
            return f


def do_if_nonexistent_file(*dargs, **kwargs):
    force = kwargs.get('force', False)
    vlevel = kwargs.get('verbose', 20)

    def wrapper(func):
        def wrapper(*args, **kwargs):
            if force:
                return func(*args, **kwargs)

            ins_a, _, _, d = inspect.getargspec(func)
            do_func = False
            checked_fns = []
            for a in dargs:
                if a in kwargs:
                    fn = kwargs[a]
                else:
                    try:
                        iarg = func.func_code.co_varnames.index(a)
                        fn = args[iarg]
                    except (IndexError, ValueError):
                        try:
                            la, ld = len(ins_a), len(d)
                            i = ins_a[la - ld:].index(a)
                        except (IndexError, ValueError):
                            msg = 'Error when defining decorator '\
                                'do_if_nonexistent_file: '\
                                '"%s" is not a valid '\
                                'argument of func %s' \
                                % (a, func.func_name)
                            raise Exception(msg)
                        fn = d[i]
                if not isinstance(fn, str):
                    raise Exception('Arg %s should be a string, %s found'
                                    % (fn, type(fn)))
                if not op.exists(fn):
                    logger.log(vlevel, 'func %s executed because file "%s" does'
                               ' not exist', func.func_name, fn)

                    do_func = True
                    break
                checked_fns.append(fn)
            if do_func:
                return func(*args, **kwargs)
            logger.log(vlevel, 'func %s not executed because file(s) '
                       'exist(s)', func.func_name)
            logger.log(vlevel, '\n'.join(['-> ' + f for f in checked_fns]))
            return None
        return wrapper
    return wrapper


def cartesian(*sequences):
    """
    Generate the "cartesian product" of all 'sequences'.  Each member of the
    product is a list containing an element taken from each original sequence.

    Note: equivalent to itertools.product, which is at least 2x faster !!

    """
    length = len(sequences)
    if length < 5:
        # Cases 1, 2 and 3, 4 are for speed only, these are not really
        # required.
        if length == 4:
            for first in sequences[0]:
                for second in sequences[1]:
                    for third in sequences[2]:
                        for fourth in sequences[3]:
                            yield [first, second, third, fourth]
        elif length == 3:
            for first in sequences[0]:
                for second in sequences[1]:
                    for third in sequences[2]:
                        yield [first, second, third]
        elif length == 2:
            for first in sequences[0]:
                for second in sequences[1]:
                    yield [first, second]
        elif length == 1:
            for first in sequences[0]:
                yield [first]
        else:
            yield []
    else:
        head, tail = sequences[:-1], sequences[-1]
        for result in cartesian(*head):
            for last in tail:
                yield result + [last]


def cartesian_combine_args(varying_args, fixed_args=None):
    """Construct the cartesian product of varying_args and append fixed_args to it.

    Parameters
    ----------
    varying_args
        Specify varying arguments as a dict mapping arg names to iterables of arg values.
        e.g:

        .. code::

            {'my_arg1' : ['a','b','c'],
             'my_arg2' : [2, 5, 10]}

    fixed_args
        Specify constant arguments as a dict mapping arg names to arg values.
        e.g:

        .. code::

            { 'my_arg3' : ['fixed_value'] }

    Examples
    --------
    >>> from pyhrf.tools import cartesian_combine_args
    >>> vargs = {'my_arg1' : ['a','b','c'],'my_arg2' : [2, 5, 10],}
    >>> fargs = { 'my_arg3' : 'fixed_value' }
    >>> res = cartesian_combine_args(vargs, fargs)
    >>> res == \
        [{'my_arg1': 'a', 'my_arg2': 2, 'my_arg3': 'fixed_value'},
    ...  {'my_arg1': 'b', 'my_arg2': 2, 'my_arg3': 'fixed_value'},
    ...  {'my_arg1': 'c', 'my_arg2': 2, 'my_arg3': 'fixed_value'},
    ...  {'my_arg1': 'a', 'my_arg2': 5, 'my_arg3': 'fixed_value'},
    ...  {'my_arg1': 'b', 'my_arg2': 5, 'my_arg3': 'fixed_value'},
    ...  {'my_arg1': 'c', 'my_arg2': 5, 'my_arg3': 'fixed_value'},
    ...  {'my_arg1': 'a', 'my_arg2': 10, 'my_arg3': 'fixed_value'},
    ...  {'my_arg1': 'b', 'my_arg2': 10, 'my_arg3': 'fixed_value'},
    ...  {'my_arg1': 'c', 'my_arg2': 10, 'my_arg3': 'fixed_value'}]
    True
    """

    if fixed_args is None:
        fixed_args = {}

    return [dict(zip(varying_args.keys(), vp) + fixed_args.items())
            for vp in iproduct(*varying_args.values())]


def icartesian_combine_args(varying_args, fixed_args=None):
    """
    Same as cartesian_combine_args but return an iterator over the
    list of argument combinations
    """

    if fixed_args is None:
        fixed_args = {}

    return (dict(zip(varying_args.keys(), vp) + fixed_args.items())
            for vp in iproduct(*varying_args.values()))


def cartesian_apply(varying_args, func, fixed_args=None, nb_parallel_procs=1, joblib_verbose=0):
    """Apply function *func* iteratively on the cartesian product of `varying_args` with fixed args `fixed_args`.
    Produce a tree (nested dicts) mapping arg values to the corresponding evaluation of function `func`

    Parameters
    ----------
    varying_args : OrderedDict
        a dictionary mapping argument names to a list of values. The `Orderdict` is used to keep track of argument order
        in the result. *WARNING:* all argument values must be hashable
    func : function
        the function to be applied on the cartesian product of given arguments
    fixed_args : dict
        arguments that are fixed (do not enter cartesian product)

    Returns
    -------
    nested dicts : tree
        where each node is an argument value from varying args and each leaf is the result of the evaluation of the
        function. The order to the tree levels corresponds the order in the input `OrderedDict` of varying arguments.

    Examples
    --------
    >>> from pyhrf.tools import cartesian_apply
    >>> from pyhrf.tools.backports import OrderedDict
    >>> def foo(a,b,c): return a + b + c
    >>> v_args = OrderedDict( [('a',[0,1]), ('b',[1,2])] )
    >>> fixed_args = {'c': 3}
    >>> cartesian_apply(v_args, foo, fixed_args) == { 0 : { 1:4, 2:5}, 1 : { 1:5, 2:6} }
    True

    """
    from pyhrf.tools.backports import OrderedDict
    assert isinstance(varying_args, OrderedDict)

    if nb_parallel_procs == 1:
        args_iter = icartesian_combine_args(varying_args, fixed_args)
        return tree([([kwargs[a] for a in varying_args.keys()], func(**kwargs))
                     for kwargs in args_iter])
    else:
        from joblib import Parallel, delayed
        p = Parallel(n_jobs=nb_parallel_procs, verbose=joblib_verbose)
        args = cartesian_combine_args(varying_args, fixed_args)
        results = p(delayed(func)(**kwargs) for kwargs in args)
        return tree([([kwargs[a] for a in varying_args.keys()], r)
                     for kwargs, r in zip(args, results)])


def format_duration(dt):
    s = ''
    if dt / 3600 >= 1:
        s += '%dH' % int(dt / 3600)
        dt = dt % 3600
    if dt / 60 >= 1:
        s += '%dmin' % int(dt / 60)
        dt = int(dt % 60)
    s += '%1.3fsec' % dt
    return s

import pyhrf.ndarray


def swapaxes(array, a1, a2):

    if isinstance(array, np.ndarray):
        return np.swapaxes(array, a1, a2)
    elif isinstance(array, pyhrf.ndarray.xndarray):
        return array.swapaxes(a1, a2)
    else:
        raise Exception('Unknown array type: %s' % str(type(array)))


def rescale_values(a, v_min=0., v_max=1., axis=None):
    a = a.astype(np.float64)  # make sure that precision is sufficient
    a_min = a.min(axis=axis)
    a_max = a.max(axis=axis)

    if axis is not None and axis != 0:
        # make target axis be the 1st to enable bcast
        a = np.swapaxes(a, 0, axis)

    res = (v_min - v_max) * 1. / (a_min - a_max) * (a - a_max) + v_max
    if axis is not None and axis != 0:
        # reposition target axis at original pos
        res = np.swapaxes(res, 0, axis)
    return res


def cartesian_params(**kwargs):
    keys = kwargs.keys()
    for p in cartesian(*kwargs.itervalues()):
        yield dict(zip(keys, p))


def cartesian_eval(func, varargs, fixedargs=None):
    resultTree = {}
    if fixedargs is None:
        fixedargs = {}
    for p in cartesian_params(**varargs):
        fargs = dict(p.items() + fixedargs.items())
        set_leaf(resultTree, [p[k] for k in varargs.iterkeys()], func(**fargs))
    return varargs.keys(), resultTree


def cuboidPrinter(c):
    print c.descrip()


def my_func(**kwargs):
    from pyhrf.ndarray import xndarray
    return xndarray(np.zeros(kwargs['shape']) + kwargs['val'])


def cartesian_test():
    branchLabels, resTree = cartesian_eval(my_func, {'shape': [(2, 5), (6, 8)],
                                                     'val': [4, 1.3]})
    pprint(resTree)
    apply_to_leaves(resTree, cuboidPrinter)


def crop_array(a, m=None, extension=0):
    """
    Return a sub array where as many zeros as possible are discarded
    Increase bounding box of mask by *extension*
    """
    if m is None:
        m = np.where(a != 0.)
    else:
        m = np.where(m != 0)
    mm = np.vstack(m).transpose()
    d = np.zeros(tuple(mm.ptp(0) + 1 + 2 * extension), dtype=a.dtype)
    d[tuple((mm - mm.min(0) + extension).transpose())] = a[m]
    return d


def buildPolyMat(paramLFD, n, dt):

    regressors = dt * np.arange(0, n)
    timePower = np.arange(0, paramLFD + 1, dtype=int)
    regMat = np.zeros((len(regressors), paramLFD + 1), dtype=float)
    logger.info('regMat: %s', str(regMat.shape))
    for v in xrange(paramLFD + 1):
        regMat[:, v] = regressors[:]

    tPowerMat = np.tile(timePower, (n, 1))
    lfdMat = np.power(regMat, tPowerMat)
    lfdMat = np.array(scipy.linalg.orth(lfdMat))
    # print 'lfdMat :', lfdMat
    return lfdMat


def polyFit(signal, tr=1., order=5):
    """
    Polynomial fit of signals.
    'signal' is a 2D matrix with first axis being time and second being position.
    'tr' is the time resolution (dt).
    'order' is the order of the polynom.
    Return the orthogonal polynom basis matrix (P) and fitted coefficients (l),
    such that P.l yields fitted polynoms.
    """
    n = len(signal)
    print 'n:', n, 'tr:', tr
    p = buildPolyMat(order, n, tr)
    ptp = np.dot(p.transpose(), p)
    invptp = np.linalg.inv(ptp)
    invptppt = np.dot(invptp, p.transpose())
    l = np.dot(invptppt, signal)
    return (p, l)


def undrift(signal, tr, order=5):
    """
    Remove the low frequency trend from 'signal' by a polynomial fit.
    Assume axis 3 of 'signal' is time.
    """
    print 'signal:', signal.shape
    m = np.where(np.ones(signal.shape[:3]))
    sm = string.join(['m[%d]' % d for d in xrange(signal.ndim - 1)], ',')
    signal_flat = eval('signal[%s,:]' % sm)
    print 'signal_flat:', signal_flat.shape
    # Least square estimate of drift
    p, l = polyFit(signal_flat.transpose(), tr, order)
    usignal_flat = signal_flat.transpose() - np.dot(p, l)
    usignal = np.zeros_like(signal)
    print 'usignal:', usignal.shape
    exec('usignal[%s,:] = usignal_flat.transpose()' % sm)
    return usignal


def root3(listCoeffs):
    length = len(listCoeffs)
    if length != 4:
        raise polyError(listCoeffs, "wrong poly order")
    if listCoeffs[0] == 0:
        raise polyError(listCoeffs[0], "wrong poly order:null coefficient")
    a = P[1] / P[0]
    b = P[2] / P[0]
    c = P[2] / P[0]

    # change of variables: z = x -a/3
    # Polynome Q(Z)=Z^3 - pZ -q
    p = a ** 2 / 3. - b
    q = (a * b) / 3. - 2. / 27. * a ** 3 - c
    if np.abs(p) < 1e-16:
        polycoeffs = np.zeros((1, 3), dtype=complex)
        polycoeffs[0] = 1
        polycoeffs[1] = (1j) ** (4 / 3.)
        polycoeffs[2] = (-1j) ** (4 / 3.)
        rp = np.multiply(polycoeffs, (np.sign(q) * q) ** (1 / 3.)) - a / 3.
    elif p < 0:
        t2 = 2 * p / 3. / q * np.sqrt(-p / 3.)
        tho = ((np.sqrt(1. + t2 ** 2) - 1) / t2) ** (1 / 3.)
        tho2 = 2. * tho / (1 - tho ** 2)
        tho3 = 2. * tho / (1 + tho ** 2)
        rp = - a / 3. * np.ones((1, 3), dtype=complex)
        fracTho2 = np.sqrt(-p / 3.) / tho2
        fracTho3 = np.sqrt(-p) / tho3
        rp[0] += -2. * fracTho2
        rp[1] += fracTho2 + 1j * fracTho3
        rp[2] += fracTho2 - 1j * fracTho3
    else:
        if np.abs((p / 3.) ** 3 - q ** 2 / 2.) < 1e-16:
            rp = - a / 3. * np.ones((1, 3), dtype=float)
            rp[0] += -3 * q / 2. / p
            rp[1] += -3 * q / 2. / p
            rp[2] += 3. * q / p


def gaussian_kernel(shape):
    """ Returns a normalized ND gauss kernel array for convolutions """
    grid = eval(
        'np.mgrid[%s]' % (string.join(['-%d:%d+1' % (s, s) for s in shape], ',')))
    k = 1. / np.prod([np.exp((grid[d] ** 2 / float(shape[d])))
                      for d in xrange(len(shape))], axis=0)
    return k / k.sum()


def gaussian_blur(a, shape):
    assert a.ndim == len(shape)
    k = gaussian_kernel(shape)
    return scipy.signal.convolve(a, k, mode='valid')


def foo(*args, **kwargs):
    pass


class polyError(Exception):
    def __init__(self, expression, message):
        self.expression = expression
        self.message = message


def convex_hull_mask(mask):
    """
    Compute the convex hull of the positions defined by the given binary mask

    Args:
        - mask (numpy.ndarray): binary mask of positions to build the chull from

    Return:
        a numpy.ndarray binary mask of positions within the convex hull

    """
    from scipy.spatial import Delaunay

    hull = Delaunay(np.array(np.where(mask)).T)

    result = np.zeros_like(mask)
    m = np.where(np.ones_like(mask))
    result[m] = hull.find_simplex(np.array(m).T) >= 0

    return result


def peelVolume3D(volume, backgroundLabel=0):

    # Make sure that boundaries are filled with zeros
    # -> expand volume with zeros:
    vs = volume.shape
    expVol = np.zeros((vs[0] + 2, vs[1] + 2, vs[2] + 2), dtype=int)
    expVol[1:-1, 1:-1, 1:-1] = volume != backgroundLabel

    # 27-neighbourhood mask:
    neighbMask = np.array([c for c in cartesian([0, -1, 1],
                                                [0, -1, 1],
                                                [0, -1, 1])][1:])
    mask = np.where(expVol != 0)
    coords = np.array(mask).transpose()
    # For each position, compute the number of valid neighbours:
    marks = np.zeros_like(expVol)
    for iv in xrange(len(mask[0])):
        cn = (neighbMask + coords[iv]).transpose()
        curNeighbMask = (cn[0], cn[1], cn[2])
        marks[mask[0][iv], mask[1][iv], mask[2][iv]] = expVol[
            curNeighbMask].sum()

    # Let's go back to the original volume shape:
    trimmedMarks = marks[1:-1, 1:-1, 1:-1]

    # Keep only positions which have 26 neighbours (completely burried):
    peeledVolume = np.zeros_like(volume)
    validPos = np.where(trimmedMarks == 26)
    peeledVolume[validPos] = volume[validPos]

    return peeledVolume


def distance(c1, c2, coord_system=None):
    # TODO: use coordinate system (xform)
    return ((c1 - c2) ** 2).sum() ** .5


def inspect_default_args(args, defaults):
    if defaults is None:
        return {}

    kw_defaults = {}
    firstdefault = len(args) - len(defaults)
    for i in range(firstdefault, len(args)):
        kw_defaults[args[i]] = defaults[i - firstdefault]

    return kw_defaults


class Pipeline:

    THE_ROOT = 0

    def __init__(self, quantities):  # , cached=False, cache_dir='./'):
        """
        Handles a graph of quantities. A quantity can be a variable or
        a callable.
        """
        self.roots = set([])  # will store roots, ie quantities
        # which have no dependency
        self.quantities = {}  # mapping value_label => compute function
        # will hold all labels:
        self.labels = set()
        self.siblings = {}  # sibling labels associated to the same quantity
        # -> eg when a func returns multiple values

        for l, q in quantities.iteritems():
            if isinstance(l, (tuple, list)):
                for e in l:
                    self.quantities[e] = q
                    self.siblings[e] = l
                self.labels.update(l)
            else:
                self.quantities[l] = q
                self.siblings[l] = (l,)
                self.labels.add(l)

        logger.info(
            'labels at init: %s (%d)', str(self.labels), len(self.labels))
        self.dependencies = dict((l, set()) for l in self.labels)
        self.dependers = dict((l, set()) for l in self.labels)
        # virtual common root of the forest
        self.dependers[self.THE_ROOT] = set()
        self.values = {}  # will hold all values
        self.init_dependencies(quantities)

    def add_root(self, label):
        self.roots.add(label)
        # Virtually plant the forest at a common root:
        self.dependers[self.THE_ROOT].add(label)

    def init_dependencies(self, quantities):

        logger.info('Pipeline.init_dependencies ...')
        for labels, val in quantities.iteritems():
            if not isinstance(labels, (list, tuple)):
                labels = (labels,)
            logger.info('treating quantities: %s', str(labels))
            logger.info('val: %s', str(val))
            func = self.get_func(val)
            if func is not None:
                logger.info('... is a function')
                arg_spec = inspect.getargspec(func)
                args = arg_spec[0]
                for label in labels:
                    assert label not in args

                    for arg in args:
                        if self.dependers.has_key(arg):
                            self.dependencies[label].add(arg)
                            self.dependers[arg].add(label)
                        else:
                            if arg_spec[3] is None or \
                                    arg not in args[len(args) - len(arg_spec[3]):]:
                                raise Exception('arg "%s" of function "%s" '
                                                'undefined (no quantity found'
                                                ' or no default value)'
                                                % (arg, val.__name__))

                    if len(self.dependencies[label]) == 0:
                        self.add_root(label)

            else:
                logger.info('... is of type %s', val.__class__)
                for label in labels:
                    self.add_root(label)

        if 0:
            print 'dependers:'
            print self.dependers
            print 'dependencies:'
            print self.dependencies
            print 'roots:'
            print self.roots
            print 'self.quantities:'
            print self.quantities
        self.checkGraph()

    def save_graph_plot(self, image_filename, images=None):
        import pygraphviz as pgv

        g = pgv.AGraph(directed=True)
        for label in self.labels:
            for deper in self.dependers[label]:
                g.add_edge(label, deper)

        for label in self.roots:
            try:
                n = g.get_node(label)
                n.attr['shape'] = 'box'
            except Exception, e:
                print e
                pass

        if images is not None:
            blank_image = pyhrf.get_data_file_name('empty.png')
            for label in self.labels:
                n = g.get_node(label)
                if images.has_key(label):
                    n.attr['image'] = images[label]
                    n.attr['labelloc'] = 'b'
                else:
                    n.attr['image'] = blank_image

        g.layout('dot')
        g.draw(image_filename)

    def update_subgraph(self, root):
        # TODO : limit init of force_eval only to quantities involved
        #       in current subgraph
        self.force_eval = dict([(label, False) for label in self.labels])

        depths = {}
        for lab in self.labels:
            depths[lab] = -1  # mark as not visited
        self.setDepths(root, depths, 0)
        maxDepth = max(depths.values())
        levels = range(maxDepth + 1)

        levels = [[] for d in xrange(maxDepth + 1)]
        for lab, depth in depths.iteritems():
            if depth != -1:
                levels[depth].append(lab)

        updated = dict((l, False) for l in self.labels)
        for level in levels:
            for lab in level:
                self.update_quantity(lab, updated)

    def update_all(self):
        self.update_subgraph(self.THE_ROOT)

    def setDepths(self, label, depths, curDepth):
        for deper in self.dependers[label]:
            # if depth not yet set for this node
            # or
            # if a deeper path has been found to  reach this node :
            if depths[deper] == -1 or depths[deper] < curDepth:
                depths[deper] = curDepth
                self.setDepths(deper, depths, curDepth + 1)

    def resolve(self):
        self.update_subgraph(self.THE_ROOT)

    def get_value(self, label):
        """
        Return the value associated with 'label'
        """
        if len(self.values) == 0:
            self.resolve()

        return self.values[label]

    def get_values(self):
        """
        Return all computed values. Perform a full update if not done yet.
        """
        if len(self.values) == 0:
            self.resolve()

        return self.values

    def reportChange(self, rootLabel):
        """
        Trigger update of the sub graph starting at the given root
        """
        assert rootLabel in self.roots
        self.update_subgraph(rootLabel)

    def reprAllDeps(self):
        """
        Build a string representing the while graph : a concatenation of
        representations of all nodes (see reprDep)
        """
        s = ""
        for lab in self.labels:
            s += self.reprDep(lab)
            s += '\n'
        return s

    def reprDep(self, label):
        """Build a string representing all dependencies and dependers of the variable `label`. The returned string is
        in the form ::

                   label
            depee1 <-
            depee2 <-
                   -> deper1
                   -> deper2

        """
        deps = self.dependencies[label]
        if len(deps) > 0:
            maxLDepee = max([len(dep) for dep in deps])
        else:
            maxLDepee = 1
        depeeMark = ' <-\n'
        s = string.join([string.ljust(' ' + dep, maxLDepee) for dep in deps] + [''],
                        depeeMark)
        deps = self.dependers[label]
        deperMark = string.rjust('-> ', maxLDepee + len(depeeMark) + 1)
        s += string.join([deperMark + dep for dep in deps], '\n')
        res = '*' + string.rjust(label, maxLDepee) + '*\n' + s
        return res

    def checkGraph(self):
        """
        Check the rightness of the builded graph (acyclicity, uniqueness and no
        short-circuits)
        """

        # check acyclicity :
        self.visited = {}
        for r in self.roots:
            self.detectCyclity([r])

        # Check there is no short-circuit:
        depths = {}
        for r in self.roots:
            for lab in self.labels:
                depths[lab] = -1
            depths[r] = 0
            self.detectShortCircuit(r, 1, depths)

    def detectShortCircuit(self, curRoot, curDepth, depths):
        """
        Recursive method which detects and corrects short-circuits
        """
        # Breadth graph walk :
        for deper in self.dependers[curRoot]:
            dDeper = depths[deper]
            # if depender was visited and its depth is smaller
            if dDeper != -1 and dDeper < curDepth:
                # Short-circuit detected -> removing it :
                self.removeShortCircuits(deper, depths)
            # Setting depth to the right one :
            depths[dDeper] = curDepth
        # Continue walking ...
        for deper in self.dependers[curRoot]:
            self.detectShortCircuit(deper, curDepth + 1, depths)

    def removeShortCircuits(self, label, depths):
        d = depths[label]
        # Look in direction leaf -> root, one level :
        for depee in self.dependencies[label]:
            dDepee = depths[depee]
            # if dependence was not visited and dependence if further than
            # depth-1 :
            if dDepee != -1 and dDepee != d - 1:
                # Remove discovered shunt :
                print 'removing shunt : %s <-> %s' \
                    % (depee, label)
                self.dependers[depee].remove(label)
                self.dependencies[label].remove(depee)

    def detectCyclity(self, viewedNodes):
        # Depth graph walk, root->leaf direction :
        # if viewedNodes
        root = viewedNodes[-1]
        for deper in self.dependers[root]:
            if deper in viewedNodes:
                msg = 'Cyclicity detected in dependency graph :\n' + \
                    ' origin is %s and dep is %s' % (root, deper)
                raise Exception(msg)
            else:
                viewedNodes.append(deper)
                self.detectCyclity(viewedNodes)
        viewedNodes.pop()

    def get_func(self, f):
        try:
            from joblib.memory import MemorizedFunc
        except ImportError:
            class MemorizedFunc:
                pass  # Dummy class

        if isinstance(f, MemorizedFunc):
            return f.func
        elif inspect.isfunction(f):
            return f
        else:
            return None

    def update_quantity(self, label, updated):

        if updated[label]:  # already updated
            return

        logger.info(" ------------- Update quantity '%s' -----------", label)
        quantity = self.quantities[label]
        siblings = self.siblings[label]
        func = self.get_func(quantity)
        if func is not None:

            logger.info(" -> %s", str(func))

            t0 = time()
            fargs = {}
            args, _, _, d = inspect.getargspec(func)
            defaults = inspect_default_args(args, d)
            for depee in args:
                fargs[depee] = self.values.get(
                    depee, defaults.get(depee, None))

            logger.info('Eval of %s ...', label)
            results = quantity(**fargs)

            logger.info('Quantity %s updated in %s', label,
                        format_duration(time() - t0))
        else:
            if isinstance(quantity, np.ndarray):
                logger.info(' -> ndarray of shape %s and type %s',
                            str(quantity.shape), str(quantity.dtype))
            else:
                logger.info(" -> %s", str(quantity))
            results = quantity

        if len(siblings) > 1:
            assert len(results) == len(siblings)

        else:
            results = (results,)

        for l, r in zip(siblings, results):
            self.values[l] = r
            updated[l] = True


def rebin(a, newshape):
    '''Rebin an array to a new shape.
    Can be used to rebin a func image onto a anat image
    '''
    assert len(a.shape) == len(newshape)

    slices = [slice(0, old, float(old) / new)
              for old, new in zip(a.shape, newshape)]
    coordinates = np.mgrid[slices]

    # choose the biggest smaller integer index:
    indices = coordinates.astype('i')

    return a[tuple(indices)]


def resampleToGrid(x, y, xgrid):

    i = 1
    yg = np.empty(xgrid.shape, dtype=y.dtype)
    for ig, xg in enumerate(xgrid):
        while i < len(x) and xg > x[i]:
            i += 1
        if i >= len(x):
            i -= 1
        dx = 0. if x[i] != x[i - 1] else 0.0000001
        yg[ig] = (y[i] - y[i - 1]) * (xg - x[i]) / \
            ((x[i] + dx) - (x[i - 1] - dx)) + y[i]
    return yg


def resampleSignal(s, osf):

    ls = len(s)
    timeMarks = np.arange(osf * (ls - 1), dtype=float) / osf
    prevTMarksSrc = np.array(np.floor(timeMarks), dtype=int)
    nextTMarksSrc = np.array(np.ceil(timeMarks), dtype=int)
    deltaBold = s[nextTMarksSrc, :] - s[prevTMarksSrc, :]
    deltaTime = timeMarks - prevTMarksSrc
    sr = s[prevTMarksSrc, :] + \
        (deltaBold.transpose() * deltaTime).transpose()
    return sr


def diagBlock(mats, rep=0):
    """
    Construct a diagonal block matrix from blocks which can be 1D or
    2D arrays. 1D arrays are taken as column vectors.
    If 'rep' is a non null positive number then blocks are diagonaly 'rep' times
    """
    if type(mats) == np.ndarray:
        finalMat = mats
    elif type(mats) == list and len(mats) == 1:
        finalMat = mats[0]
    elif type(mats) == list:
        m = 0  # nbLines
        n = 0  # nbCols
        for mat in mats:
            m += mat.shape[0]
            n += 1 if len(mat.shape) < 2 else mat.shape[1]
        finalMat = np.zeros((m, n), dtype=float)
        lOffset = 0
        cOffset = 0
        for mat in mats:
            m = mat.shape[0]
            n = 1 if len(mat.shape) < 2 else mat.shape[1]
            sh = finalMat[lOffset:lOffset + m, cOffset:cOffset + n].shape
            finalMat[
                lOffset:lOffset + m, cOffset:cOffset + n] = mat.reshape(sh)[:]
            lOffset += m
            cOffset += n
    else:
        raise Exception('diagBlock: unrecognised type for "mats"')

    if rep > 0:
        return diagBlock([finalMat] * rep)
    else:
        return finalMat


def extractRoiMask(nmask, roiId):

    mask = nmask == roiId
    m = np.where(mask)
    mm = np.vstack(m).transpose()
    cropMask = np.zeros(mm.ptp(0) + 1, dtype=int)
    cropMask[tuple((mm - mm.min(0)).transpose())] = mask[m]

    return cropMask


def describeRois(roiMask):

    s = 'Number of voxels : %d\n' % (roiMask != 0).sum()
    s += 'Number of regions : %d\n' % (
        len(np.unique(roiMask)) - int(0 in roiMask))
    s += 'Region sizes :\n'
    counts = np.bincount(roiMask[roiMask > 0])
    s += np.array2string(counts) + '\n'
    nbr = (counts >= 60).sum()
    s += 'Nb of region with size > 60: %d\n' % nbr

    return s


def array_summary(a, precision=4):
    return '%s -- %1.*f(%1.*f)[%1.*f...%1.*f]' % (str(a.shape),
                                                  precision, a.mean(),
                                                  precision, a.std(),
                                                  precision, a.min(),
                                                  precision, a.max())


def get_2Dtable_string(val, rownames, colnames, precision=4, col_sep='|',
                       line_end='', line_start='', outline_char=None):
    """ Return a nice tabular string representation of a 2D numeric array
    #TODO : make colnames and rownames optional
    """

    if val.ndim == 1:
        val = val.reshape(val.shape[0], 1)
    nrows, ncols = val.shape[:2]

    if np.isscalar(val[0, 0]):
        valWidth = len(str('%1.*f' % (precision, -3.141658938325)))
    else:
        if (val >= 0).all():
            valWidth = len(array_summary(val[0, 0], precision=precision))
        else:
            valWidth = len(array_summary(np.abs(val[0, 0]) * -1,
                                         precision=precision))

    rowWidth = max([len(str(rn)) for rn in rownames])
    colWidth = [max([valWidth, len(cn)]) for cn in colnames]

    sheader = line_start + ' ' * rowWidth + '  ' + col_sep
    sheader += col_sep.join([' %*s ' % (colWidth[j], cn)
                             for j, cn in enumerate(colnames)])
    sheader += line_end + '\n'

    scontent = ''
    for i in xrange(nrows):
        line = line_start + ' %*s ' % (rowWidth, str(rownames[i])) + col_sep
        for j in xrange(ncols):
            if np.isscalar(val[i, j]):
                line += ' %*.*f ' % (colWidth[j],
                                     precision, val[i, j]) + col_sep
            else:
                line += ' %*s ' % (valWidth,
                                   array_summary(val[i, j], precision)) + col_sep

        if outline_char is not None:
            outline = outline_char * (len(line) - 1 + len(line_end)) + '\n'
        else:
            outline = ''
        scontent += outline + line[:-len(col_sep)] + line_end + '\n'

    return outline + sheader + scontent + outline


def get_leaf(element, branch):
    """
    Return the nested leaf element corresponding to all dictionnary keys in
    branch from element
    """
    if isinstance(element, dict) and len(branch) > 0:
        return get_leaf(element[branch[0]], branch[1:])
    else:
        return element


def set_leaf(tree, branch, leaf, branch_classes=None):
    """
    Set the nested *leaf* element corresponding to all dictionnary keys
    defined in *branch*, within *tree*
    """
    assert isinstance(tree, dict)
    if len(branch) == 1:
        tree[branch[0]] = leaf
        return
    if not tree.has_key(branch[0]):
        if branch_classes is None:
            tree[branch[0]] = tree.__class__()
        else:
            tree[branch[0]] = branch_classes[0]()
    else:
        assert isinstance(tree[branch[0]], dict)
    if branch_classes is not None:
        set_leaf(tree[branch[0]], branch[1:], leaf, branch_classes[1:])
    else:
        set_leaf(tree[branch[0]], branch[1:], leaf)


def swap_layers(t, labels, l1, l2):
    """ Create a new tree from t where layers labeled by l1 and l2 are swapped.
    labels contains the branch labels of t.
    """
    i1 = labels.index(l1)
    i2 = labels.index(l2)
    nt = t.__class__()  # new tree init from the input tree
    # can be dict or OrderedDict
    for b, l in izip(treeBranches(t), tree_leaves(t)):
        nb = list(b)
        nb[i1], nb[i2] = nb[i2], nb[i1]
        set_leaf(nt, nb, l)

    return nt


def tree_rearrange(t, oldlabels, newlabels):
    """ Create a new tree from t where layers are rearranged following newlabels.
    oldlabels contains the branch labels of t.
    """
    order = [oldlabels.index(nl) for nl in newlabels]
    nt = t.__class__()  # new tree
    for b, l in izip(treeBranches(t), tree_leaves(t)):
        nb = [b[i] for i in order]
        set_leaf(nt, nb, l)

    return nt


def treeBranches(tree, branch=None):
    if branch is None:
        branch = []
    if isinstance(tree, dict):
        for k in tree.iterkeys():
            for b in treeBranches(tree[k], branch + [k]):
                yield b
    else:
        yield branch


def tree(branched_leaves):
    d = {}
    for branch, leaf in branched_leaves:
        set_leaf(d, branch, leaf)
    return d


def treeBranchesClasses(tree, branch=None):
    if branch is None:
        branch = []
    if isinstance(tree, dict):
        for k, v in tree.iteritems():
            for b in treeBranchesClasses(tree[k], branch + [v.__class__]):
                yield b
    else:
        yield branch


def tree_leaves(tree):
    for branch in treeBranches(tree):
        yield get_leaf(tree, branch)


def tree_items(tree):
    """
    """
    for branch in treeBranches(tree):
        yield (branch, get_leaf(tree, branch))


def stack_trees(trees, join_func=None):
    """ Stack trees (python dictionnaries) with identical structures
    into one tree
    so that one leaf of the resulting tree is a list of the corresponding leaves
    across input trees. 'trees' is a list of dict
    """
    stackedTree = trees[0].__class__()

    for branch, branch_classes in izip(treeBranches(trees[0]),
                                       treeBranchesClasses(trees[0])):
        if join_func is None:
            leaveList = [get_leaf(tree, branch) for tree in trees]
        else:
            leaveList = join_func([get_leaf(tree, branch) for tree in trees])
        set_leaf(stackedTree, branch, leaveList, branch_classes)
    return stackedTree


def unstack_trees(tree):
    """ Return a list of tree from a tree where leaves are all lists with
    the same number of items
    """

    first_list = tree_leaves(tree).next()
    tree_list = [tree.__class__() for i in xrange(len(first_list))]
    for b, l in tree_items(tree):
        for t, item in zip(tree_list, l):
            set_leaf(t, b, item)

    return tree_list


from pprint import pprint


def apply_to_leaves(tree, func, funcArgs=None, funcKwargs=None):
    """
    Apply function 'func' to all leaves in given 'tree' and return a new tree.
    """
    if funcKwargs is None:
        funcKwargs = {}
    if funcArgs is None:
        funcArgs = []

    newTree = tree.__class__()  # could be dict or OrderedDict
    for branch, leaf in tree_items(tree):
        set_leaf(newTree, branch, func(leaf, *funcArgs, **funcKwargs))
    return newTree


def map_dict(func, d):
    return d.__class__((k, func(v)) for k, v in d.iteritems())


def get_cache_filename(args, path='./', prefix=None, gz=True):
    hashArgs = hashlib.sha512(repr(args)).hexdigest()
    if prefix is not None:
        fn = os.path.join(path, prefix + '_' + hashArgs + '.pck')
    else:
        fn = os.path.join(path, hashArgs + '.pck')
    if gz:
        return fn + '.gz'
    else:
        return fn


def hash_func_input(func, args, digest_code):

    if isinstance(args, dict):  # sort keys
        to_digest = ''
        for k in sorted(args.keys()):
            v = args[k]
            to_digest += repr(k) + repr(v)
    else:
        to_digest = repr(args)

    if digest_code:
        to_digest += inspect.getsource(func)

    return hashlib.sha512(to_digest).hexdigest()


def cache_filename(func, args=None, prefix=None, path='./',
                   digest_code=False):
    if prefix is None:
        prefix = func.__name__
    else:
        prefix = prefix + '_' + func.__name__

    hashArgs = hash_func_input(func, args, digest_code)

    fn = os.path.join(path, prefix + '_' + hashArgs + '.pck.gz')
    return fn


def cache_exists(func, args=None, prefix=None, path='./',
                 digest_code=False):

    return op.exists(cache_filename(func, args=args, prefix=prefix,
                                    path=path, digest_code=digest_code))


def cached_eval(func, args=None, new=False, save=True, prefix=None,
                path='./', return_file=False, digest_code=False,
                gzip_mode='cmd'):

    fn = cache_filename(func, args=args, prefix=prefix,
                        path=path, digest_code=digest_code)

    if not os.path.exists(fn) or new:
        if args is None:
            r = func()
        elif isinstance(args, tuple) or isinstance(args, list):
            r = func(*args)
        elif isinstance(args, dict):
            r = func(**args)
        else:
            raise Exception("type of arg (%s) is not valid. Should be tuple, "
                            "list or dict" % str(args.__class__))
        if save:
            if gzip_mode == 'pygzip' and os.system('gzip -V') != 0:
                f = gzip.open(fn, 'w')
                cPickle.dump(r, f)
                f.close()
            else:
                f = open(fn[:-3], 'w')
                cPickle.dump(r, f)
                f.close()
                os.system("gzip -f %s" % fn[:-3])
        if not return_file:
            return r
        else:
            return fn
    else:
        if not return_file:
            return cPickle.load(gzip.open(fn))
        else:
            return fn


def montecarlo(datagen, festim, nbit=None):
    """Perform a Monte Carlo loop with data generator 'datagen' and estimation
    function 'festim'.
    'datagen' have to be iterable.
    'festim' must return an object on which ** and + operators can be applied.
    If 'nbit' is provided then use it for the maximum iterations else loop until
    datagen stops.
    """
    itgen = iter(datagen)

    s = itgen.next()
    e = festim(s)
    cumul = e
    cumul2 = e ** 2

    if nbit is None:
        nbit = 0
        for s in itgen:
            e = festim(s)
            cumul += e
            cumul2 += e ** 2
            nbit += 1
    else:
        for i in xrange(1, nbit):
            s = itgen.next()
            e = festim(s)
            cumul += e
            cumul2 += e ** 2

    m = cumul / float(nbit)
    v = cumul2 / float(nbit) - m ** 2
    return m, v


def closestsorted(a, val):
    i = np.searchsorted(a, val)
    if i == len(a) - 1:
        return i
    elif np.abs(a[i] - val) < np.abs(a[i + 1] - val):
        return i
    else:
        return i + 1


##################
def calc_nc2D(a, b):
    return a + b + 2 * (a - 1) * (b - 1) - 2


def nc2DGrid(maxSize):
    nc2D = np.zeros((maxSize, maxSize), dtype=int)
    for a in xrange(maxSize):
        for b in xrange(maxSize):
            nc2D[a, b] = calc_nc2D(a, b)
    return nc2D


def now():
    return datetime.datetime.fromtimestamp(time())


def time_diff_str(diff):
    return '%dH%dmin%dsec' % (diff.seconds / 3600,
                              (diff.seconds % 3600) / 60,
                              (diff.seconds % 3600) % 60)


class AnsiColorizer:
    """ Format strings with an ANSI escape sequence to encode color """

    BEGINC = '\033['
    COLORS = {
        'purple': '95',
        'blue': '94',
        'green': '92',
        'yellow': '93',
        'red': '91',
    }

    ENDC = '\033[0m'

    def __init__(self):
        self.disabled = False
        self.do_tty_check = True

    def disable(self):
        self.disabled = True

    def enable(self):
        self.disabled = False

    def no_tty_check(self):
        self.do_tty_check = False

    def tty_check(self):
        self.do_tty_check = True

    def __call__(self, s, color, bright=False, bold=False):
        if color not in self.COLORS:
            raise Exception('Invalid color "%s". Available colors: %s'
                            % (color, str(self.COLORS)))

        col = self.COLORS[color]
        if self.disabled or (self.do_tty_check and not sys.stdout.isatty()):
            return s
        else:
            ansi_codes = ";".join([['', '1'][bright], col])
            return '%s%sm%s%s' % (self.BEGINC, ansi_codes, s, self.ENDC)


colorizer = AnsiColorizer()


def extract_file_series(files):
    """
    group all file names sharing a common prefix followed by a number, ie:
    <prefix><number><extension>
    Return a dictionnary with two levels (<tag>,<extension>), mapped to all
    corresponding series index found.
    """
    series = {}  # will map (tag,extension) to number series
    rexpSeries = re.compile("(.*?)([0-9]+)[.]([a-zA-Z.~]*?)\Z")
    series = defaultdict(lambda: defaultdict(list))
    for f in files:
        r = rexpSeries.match(f)
        if r is not None:
            (tag, idx, ext) = r.groups()
            ext = '.' + ext
        else:
            tag, ext = op.splitext(f)
            idx = ''
        series[tag][ext].append(idx)
    return series


def format_serie(istart, iend):
    return colorizer(['[%d...%d]' % (istart, iend),
                      '[%d]' % (istart)][istart == iend],
                     'red')


def condense_series(numbers):
    if len(numbers) == 1:
        return numbers[0]
    else:
        inumbers = np.sort(np.array(map(int, numbers)))
        if (np.diff(inumbers) == 1).all():
            return format_serie(inumbers.min(), inumbers.max())
        else:
            segment_start = 0
            s = ''
            for segment_end in np.where(np.diff(inumbers) != 1)[0]:
                s += format_serie(inumbers[segment_start],
                                  inumbers[segment_end])
                segment_start = segment_end + 1
            s += format_serie(inumbers[segment_start], inumbers[-1])
            return s


def group_file_series(series, group_rules=None):
    if group_rules is None:
        group_rules = []

    groups = defaultdict(dict)
    dummy_tag = 0
    for tag, ext_data in series.iteritems():
        tag_has_been_grouped = False
        for gr in group_rules:
            r = gr.match(tag)
            if r is not None:
                gname = r.group('group_name')
                groups[gname]['...'] = gname
                tag_has_been_grouped = True
                break

        if not tag_has_been_grouped:
            for ext, numbers in ext_data.iteritems():

                if '' in numbers:
                    groups[dummy_tag][ext] = tag
                    numbers.remove('')

                if len(numbers) > 0:
                    groups[tag][ext] = tag + condense_series(numbers)
            dummy_tag += 1

    final_groups = []
    for tag, series in groups.iteritems():
        sv = series.values()
        if len(sv) > 1 and len(set(sv)) == 1:
            exts = [ext[1:] for ext in series.keys()]
            final_groups.append(
                sv[0] + colorizer('.{%s}' % ','.join(exts), 'green'))
        else:
            for ext, s in series.iteritems():
                if ext == '...':
                    ext = colorizer('...', 'purple')
                final_groups.append(s + ext)

    return sorted(final_groups)


def check_files_series(fseries, verbose=False):

    ok_status = True
    for tag, dext in fseries.iteritems():
        for ext, indexes in dext.iteritems():
            if 0:
                print 'tag', tag, 'ext', ext
                print 'indexes:', indexes
            sorted_indexes = sorted(indexes)
            last = int(sorted_indexes[-1])
            first = int(sorted_indexes[0])
            diff = last - first + 1 - len(indexes)
            if diff != 0:
                ok_status = False
                if verbose:
                    print '%d items missing for series %s[...].%s' \
                        % (diff, tag, ext)
                    print '-> Series should have %d items (%s to %s)' \
                          ' - found %d items' \
                          % (last - first + 1, first, last, len(indexes))
    return ok_status


# Factorisation of the code: functions creations
def Extract_TTP_whM_hrf(hrf, dt):
    """
    Extract TTP and whM from an hrf
    """
    from scipy.interpolate import interp1d
    from pyhrf.boldsynth.hrf import getCanoHRF
    hcano = getCanoHRF()

    Abscisses = np.arange(hrf.size) * dt

    # TTP calculus
    TTP = hrf.argmax() * dt
    print 'found TTP:', TTP

    # whM calculus
    # 1/ Round the HRF
    HRF_rounded = np.round(hrf, 5)

    # 2/ Interpolation to obtain more values
    Abscisses_round = np.arange(HRF_rounded.size) * dt
    f = interp1d(Abscisses, hrf)
    r = 0.00001
    HRF_interp = f(np.arange(0, Abscisses_round[len(Abscisses_round) - 1], r))
    HRF_interp_rounded = np.round(HRF_interp, 5)

    # To reconvert from interpolation to correct values in seconds
    len_use = len(HRF_interp)
    dt_interp = (hcano[0][len(hcano[0]) - 1]) / len(HRF_interp)

    # 3/ Where the half max is found
    Pts_1_2_h = np.where(
        abs(HRF_interp_rounded - HRF_rounded.max() / 2.) < 0.001)
    Values_pts_1_2_h = HRF_interp[Pts_1_2_h]
    if Pts_1_2_h[0].shape[0] == 0:
        print '#### No point found ####'

    # Selection of Pts of abscisse<max
    Diff1 = abs(Pts_1_2_h - HRF_interp_rounded.argmax()) * \
        (Pts_1_2_h < HRF_interp_rounded.argmax())
    Diff1_non_zeros = Diff1[0][np.where(Diff1[0] > 0)]  # retrieve positions#0
    Diff1_non_zeros.sort()  # to sort all differences
    First_diff1 = Diff1_non_zeros.mean()

    # Selection of Pts of abscisse>max
    Diff2 = abs(HRF_interp_rounded.argmax() - Pts_1_2_h) * \
        (Pts_1_2_h > HRF_interp_rounded.argmax())
    Diff2_non_zeros = Diff2[0][np.where(Diff2[0] > 0)]  # retrieve positions#0
    Diff2_non_zeros.sort()  # to sort all differences
    First_diff2 = Diff2_non_zeros.mean()

    # addition of the two differences and *dt_interp to obtain whM in seconds
    whM = (First_diff1 + First_diff2) * dt_interp
    print 'found whM:', whM

    return TTP, whM


def Extract_TTP_whM_from_group(hrfs_pck_file, dt, model, Path_data, acq):
    """
    Extract TTP and whM from a group of hrfs whose values are saved in a .pck (size nb_subjects * nb_coeff_hrf)
    """
    from scipy.interpolate import interp1d
    from pyhrf.boldsynth.hrf import getCanoHRF
    hcano = getCanoHRF()

    hrfs = cPickle.load(open(hrfs_pck_file))

    nb_subjects = hrfs.shape[0]

    TTP_tot = np.zeros((nb_subjects))
    whM_tot = np.zeros((nb_subjects))

    for isubj in np.arange(nb_subjects):

        HRF_at_max = hrfs[isubj, :]

        Abscisses = np.arange(HRF_at_max.size) * dt

        # TTP calculus
        TTP = HRF_at_max.argmax() * dt
        print 'found TTP:', TTP
        TTP_tot[isubj] = TTP

        # whM calculus
        # 1/ Round the HRF
        HRF_rounded = np.round(HRF_at_max, 5)

        # 2/ Interpolation to obtain more values
        Abscisses_round = np.arange(HRF_rounded.size) * dt
        f = interp1d(Abscisses, HRF_at_max)
        r = 0.00001
        HRF_interp = f(
            np.arange(0, Abscisses_round[len(Abscisses_round) - 1], r))
        HRF_interp_rounded = np.round(HRF_interp, 5)

        # To reconvert from interpolation to correct values in seconds
        len_use = len(HRF_interp)
        dt_interp = (hcano[0][len(hcano[0]) - 1]) / len(HRF_interp)

        # 3/ Where the half max is found
        Pts_1_2_h = np.where(
            abs(HRF_interp_rounded - HRF_rounded.max() / 2.) < 0.001)
        Values_pts_1_2_h = HRF_interp[Pts_1_2_h]
        if Pts_1_2_h[0].shape[0] == 0:
            print '#### No point found ####'

        # Selection of Pts of abscisse<max
        Diff1 = abs(Pts_1_2_h - HRF_interp_rounded.argmax()) * \
            (Pts_1_2_h < HRF_interp_rounded.argmax())
        # retrieve positions#0
        Diff1_non_zeros = Diff1[0][np.where(Diff1[0] > 0)]
        Diff1_non_zeros.sort()  # to sort all differences
        First_diff1 = Diff1_non_zeros.mean()

        # Selection of Pts of abscisse>max
        Diff2 = abs(HRF_interp_rounded.argmax() - Pts_1_2_h) * \
            (Pts_1_2_h > HRF_interp_rounded.argmax())
        # retrieve positions#0
        Diff2_non_zeros = Diff2[0][np.where(Diff2[0] > 0)]
        Diff2_non_zeros.sort()  # to sort all differences
        First_diff2 = Diff2_non_zeros.mean()

        # addition of the two differences and *dt_interp to obtain whM in
        # seconds
        whM = (First_diff1 + First_diff2) * dt_interp
        print 'found whM:', whM

        whM_tot[isubj] = np.round(whM, 1)

        cPickle.dump(TTP_tot, open(
            Path_data + '/_TTPs_at_peak_by_hand_' + '_' + model + '_' + acq + '.pck', 'w'))
        cPickle.dump(whM_tot, open(
            Path_data + '/_whMs_at_peak_by_hand_' + '_' + model + '_' + acq + '.pck', 'w'))

    return TTP_tot, whM_tot


def PPMcalculus_jde(threshold_value, apost_mean_activ_fn, apost_var_activ_fn,
                    apost_mean_inactiv_fn, apost_var_inactiv_fn, labels_activ_fn,
                    labels_inactiv_fn, nrls_fn, mask_file, null_hyp=True):
    '''
    Function to calculate the probability that the nrl in voxel j,
    condition m, is superior to a given hreshold_value
    Computation for all voxels
    Compute Tvalue according to null hypothesis
    '''
    from scipy.integrate import quad
    from pyhrf.ndarray import xndarray
    from scipy.stats import norm

    mask = xndarray.load(mask_file).data
    apost_mean_activ = xndarray.load(apost_mean_activ_fn)
    apost_mean_inactiv = xndarray.load(apost_mean_inactiv_fn)
    apost_var_activ = xndarray.load(apost_var_activ_fn)
    apost_var_inactiv = xndarray.load(apost_var_inactiv_fn)
    labels_activ = xndarray.load(labels_activ_fn)
    labels_inactiv = xndarray.load(labels_inactiv_fn)
    nrls = xndarray.load(nrls_fn)

    # flattend data
    m1 = apost_mean_activ.flatten(
        mask, axes=['sagittal', 'coronal', 'axial'], new_axis='position').data
    m2 = apost_mean_inactiv.flatten(
        mask, axes=['sagittal', 'coronal', 'axial'], new_axis='position').data
    var1 = apost_var_activ.flatten(
        mask, axes=['sagittal', 'coronal', 'axial'], new_axis='position').data
    var2 = apost_var_inactiv.flatten(
        mask, axes=['sagittal', 'coronal', 'axial'], new_axis='position').data
    perc1 = labels_activ.flatten(
        mask, axes=['sagittal', 'coronal', 'axial'], new_axis='position').data
    perc2 = labels_inactiv.flatten(
        mask, axes=['sagittal', 'coronal', 'axial'], new_axis='position').data
    nrls_values = nrls.flatten(
        mask, axes=['sagittal', 'coronal', 'axial'], new_axis='position').data

    Probas = np.zeros(perc1.shape[0])
    if null_hyp:
        Pvalues = np.zeros(perc1.shape[0])
        # to detect positions activ and inactiv
        Comp = perc1 - perc2
        Pos_activ = np.where(Comp[:, 0])
        Pos_inactiv = np.where(Comp[:, 0] < 0)

        Means = np.zeros(perc1.shape[0])
        Vars = np.zeros(perc1.shape[0])

    for i in xrange(perc1.shape[0]):
        # posterior probability distribution
        fmix = lambda x: perc1[
            i] * norm.pdf(x, m1[i], var1[i] ** .5) + perc2[i] * norm.pdf(x, m2[i], var2[i] ** .5)

        Probas[i] = quad(fmix, threshold_value, float('inf'))[0]

    if null_hyp:
        Means[Pos_activ] = m1[Pos_activ]
        Means[Pos_inactiv] = m2[Pos_inactiv]
        Vars[Pos_activ] = var1[Pos_activ]
        Vars[Pos_inactiv] = var2[Pos_inactiv]
        for i in xrange(perc1.shape[0]):
            nrl_val = nrls_values[i]
            fmix = lambda x: norm.pdf(x, 0, Vars[i] ** .5)
            Pvalues[i] = quad(fmix, nrl_val, float('inf'))[0]

    # deflatten to retrieve original shape
    PPM_ = xndarray(Probas, axes_names=['position'])
    PPM = PPM_.expand(mask, 'position', ['sagittal', 'coronal', 'axial'])

    PPMinvv = 1 - Probas  # to obtain more readable maps
    PPMinv_ = xndarray(PPMinvv, axes_names=['position'])
    PPMinv = PPMinv_.expand(mask, 'position', ['sagittal', 'coronal', 'axial'])

    Pval_ = xndarray(Pvalues, axes_names=['position'])
    Pval = Pval_.expand(mask, 'position', ['sagittal', 'coronal', 'axial'])

    return PPM.data, PPMinv.data, Pval.data


## HTML formating ##

def html_row(s):
    return '<tr>%s</tr>' % s


def html_table(s, border=None):
    if border is None:
        return '<table>%s</table>' % s
    else:
        return '<table border="%d">%s</table>' % (border, s)


def attrs_to_string(attrs):
    attrs = attrs or {}
    sattrs = ''
    if len(attrs) > 0:
        sattrs = ' ' + ' '.join(['%s="%s"' % (k, v) for k, v in attrs.items()])
    return sattrs


def html_img(fn, attrs=None):
    return '<img src="%s"%s>' % (fn, attrs_to_string(attrs))


def html_cell(s, cell_type='d', attrs=None):
    return '<t%s%s>%s</t%s>' % (cell_type, attrs_to_string(attrs), s, cell_type)


def html_div(s, attrs=None):
    return '<div%s>%s</div>' % (attrs_to_string(attrs), s)


def html_list_to_row(l, cell_types, attrs):
    if not isinstance(attrs, (list, tuple)):
        attrs = [attrs] * len(l)
    else:
        assert len(attrs) == len(l)

    if not isinstance(cell_types, (list, tuple)):
        cell_types = [cell_types] * len(l)
    else:
        assert len(cell_types) == len(l)

    return html_row(''.join([html_cell(e, t, a)
                             for e, t, a in zip(l, cell_types, attrs)]))


def html_doc(s):
    return '<!DOCTYPE html><html>' + s + '</html>'


def html_head(s):
    return '<head>' + s + '</head>'


def html_style(s):
    return '<style>' + s + '</style>'


def html_body(s):
    return '<body>' + s + '</body>'
